# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import datetime
import platform
import abc
from abc import ABC

from pyworkflow.gui.project.constants import STATUS_COLORS, WARNING_COLOR
from pyworkflow.protocol import STATUS_FAILED
from pyworkflow.viewer import ProtocolViewer


def getStatusColorFromNode(node):
    # If it is a run node (not PROJECT)
    return getStatusColorFromRun(node.run)


def getStatusColorFromRun(prot):
    """ Returns the color associated with the status. """
    if prot:
        if prot.hasSummaryWarnings():
            return WARNING_COLOR
        else:
            return getStatusColor(prot.status.get(STATUS_FAILED))
    else:
        return getStatusColor()


def getStatusColor(status=None, default='#ADD8E6'):
    """
    Parameters
    ----------
    status status of the protocol

    Returns the color associated with he status
    -------

    """
    return STATUS_COLORS[status] if status else default

# OS dependent behaviour. Add any OS dependent method here and later we might move
# or refactor this to a class or something else


class OSHandler(abc.ABC):
    """ Abstract class: Handler for OS specific actions"""
    def maximizeWindow(root):
        pass


class LinuxHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.attributes("-zoomed", True)


class MacHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.state("zoomed")


class WindowsHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.state("zoomed")


class OS:
    _handler = None

    _handlers = {"Linux": LinuxHandler,
                 "Darwin": MacHandler,
                 "Windows": WindowsHandler}  # Until testing this on windows

    @staticmethod
    def getPlatform():
        return platform.system()

    @classmethod
    def handler(cls):
        if cls._handler is None:
            cls._handler = cls._handlers[cls.getPlatform()]

        return cls._handler

    @classmethod
    def getDistro(cls):
        return platform.os.uname().version


def isAFinalProtocol(v, k):
    if (issubclass(v, ProtocolViewer) or
            v.isBase() or v.isDisabled()):
        return False

    return v.__name__ == k


def inspectObj(object, filename, prefix='', maxDeep=5, inspectDetail=2, memoryDict=None):
    """ Creates a .CSV file in the filename path with
        all its members and recursively with a certain maxDeep,
        if maxDeep=0 means no maxDeep (until all members are inspected).

        inspectDetail can be:
         - 1: All attributes are shown
         - 2: All attributes are shown and iterable values are also inspected

        prefix and memoryDict will be updated in the recursive entries:
         - prefix is a compound of the two first columns (DEEP and Tree)
         - memoryDict is a dictionary with the memory address and an identifier
    """
    END_LINE = '\n'  # end of line char
    COL_DELIM = '\t'  # column delimiter
    INDENT_COUNTER = '/'  # character append in each indention (it's not written)

    NEW_CHILD = '  |------>  '  # new item indention
    BAR_CHILD = '  | ' + INDENT_COUNTER  # bar indention
    END_CHILD = ('       -- ' + COL_DELIM) * 4 + END_LINE  # Child ending
    column1 = '    - Name - ' + COL_DELIM
    column2 = '    - Type - ' + COL_DELIM
    column3 = '    - Value - ' + COL_DELIM
    column4 = '  - Memory Address -'

    #  Constants to distinguish the first, last and middle rows
    IS_FIRST = 1
    IS_LAST = -1
    IS_MIDDLE = 0

    memoryDict = memoryDict or {}

    def writeRow(name, value, prefix, posList=False):
        """ Writes a row item. """
        # we will avoid to recursively print the items wrote before
        #  (ie. with the same memory address), thus we store a dict with the
        #  addresses and the flag isNew is properly set
        if str(hex(id(value))) in memoryDict:
            memorySTR = memoryDict[str(hex(id(value)))]
            isNew = False
        else:
            # if the item is new, we save its memory address in the memoryDict
            #   and we pass the name and the line on the file as a reference.
            memorySTR = str(hex(id(value)))
            file = open(filename, 'r')
            lineNum = str(len(file.readlines()) + 1)
            file.close()
            nameDict = str(name)[0:15] + ' ...' if len(str(name)) > 25 else str(name)
            memoryDict[str(hex(id(value)))] = '>>> ' + nameDict + ' - L:' + lineNum
            isNew = True

        if posList:
            # if we have a List, the third column is 'pos/lenght'
            thirdCol = posList
        else:
            # else, we print the value avoiding the EndOfLine char (// instead)
            thirdCol = str(value).replace(END_LINE, ' // ')

        # we will print the indentation deep number in the first row
        indentionDeep = prefix.count(INDENT_COUNTER)
        deepStr = str(indentionDeep) + COL_DELIM

        # the prefix without the indentCounters is
        #   the tree to be printed in the 2nd row
        prefixToWrite = prefix.replace(INDENT_COUNTER, '')

        file = open(filename, 'a')
        file.write(deepStr + prefixToWrite + COL_DELIM +
                   str(name) + COL_DELIM +
                   str(type(value)) + COL_DELIM +
                   thirdCol + COL_DELIM +
                   memorySTR + END_LINE)
        file.close()

        return isNew

    def recursivePrint(value, prefix, isFirstOrLast):
        """ We print the childs items of tuples, lists, dicts and classes. """

        # if it's the last item, its childs has not the bar indention
        if isFirstOrLast == IS_LAST:  # void indention when no more items
            prefixList = prefix.split(INDENT_COUNTER)
            prefixList[-2] = prefixList[-2].replace('|', ' ')
            prefix = INDENT_COUNTER.join(prefixList)

        # recursive step with the new prefix and memory dict.
        inspectObj(value, filename, prefix + BAR_CHILD, maxDeep, inspectDetail,
                   memoryDict)

        if isFirstOrLast == IS_FIRST:
            deepStr = str(indentionDeep) + COL_DELIM
        else:
            # When it was not the first item, the deep is increased
            #   to improve the readability when filter
            deepStr = str(indentionDeep + 1) + COL_DELIM

        prefix = prefix.replace(INDENT_COUNTER, '') + COL_DELIM

        # We introduce the end of the child and
        #   also the next header while it is not the last
        file = open(filename, 'a')
        file.write(deepStr + prefix + END_CHILD)
        if isFirstOrLast != IS_LAST:
            # header
            file.write(deepStr + prefix +
                       column1 + column2 + column3 + column4 + END_LINE)
        file.close()

    def isIterable(obj):
        """ Returns true if obj is a tuple, list, dict or calls. """
        isTupleListDict = (isinstance(obj, tuple) or
                           isinstance(obj, dict) or
                           isinstance(obj, list)) and len(value) > 1

        # FIX ME: I don't know how to assert if is a class or not...
        isClass = str(type(obj))[1] == 'c'

        return isClass or (isTupleListDict and inspectDetail < 2)

    indentionDeep = prefix.count(INDENT_COUNTER)
    if indentionDeep == 0:
        prefix = ' - Root - '

        # dict with name and value pairs of the members
        if len(object) == 1:
            # if only one obj is passed in the input list,
            #   we directly inspect that obj.
            obj_dict = object[0].__dict__
            object = object[0]

        #  setting the header row
        treeHeader = ' - Print on ' + str(datetime.datetime.now())
        prefixHeader = '-DEEP-' + COL_DELIM + treeHeader + COL_DELIM
        col1 = '    - Name - (value for Lists and Tuples)' + COL_DELIM
        col3 = '    - Value - (Pos./Len for Lists and Tuples) ' + COL_DELIM

        #  writing the header row
        file = open(filename, 'w')
        file.write(prefixHeader + col1 + column2 + col3 + column4 + END_LINE)
        file.close()

        #  writing the root object
        writeRow(object.__class__.__name__, object, prefix)
        #  adding the child bar to the prefix
        prefix = '  ' + BAR_CHILD
    else:
        # firsts settings depending on the type of the obj
        if str(type(object))[1] == 'c':
            obj_dict = object.__dict__
        elif (isinstance(object, tuple) or
              isinstance(object, list)):
            column1 = '    - Value - ' + COL_DELIM
            column3 = '  - Pos./Len. - ' + COL_DELIM
        elif isinstance(object, dict):
            column1 = '    - Key - ' + COL_DELIM
            obj_dict = object
        else:  # if is not of the type above it not make sense to continue
            return

    indentionDeep = prefix.count(INDENT_COUNTER)
    deepStr = str(indentionDeep) + COL_DELIM
    isBelowMaxDeep = indentionDeep < maxDeep if maxDeep > 0 else True

    prefixToWrite = prefix.replace(INDENT_COUNTER, '') + COL_DELIM
    file = open(filename, 'a')
    file.write(deepStr + prefixToWrite +
               column1 + column2 + column3 + column4 + END_LINE)
    file.close()

    #  we update the prefix to put the NEW_CHILD string  ( |----> )
    prefixList = prefix.split(INDENT_COUNTER)
    prefixList[-2] = NEW_CHILD
    #  we return to the string structure
    #    with a certain indention if it's the root
    prefixToWrite = '  ' + INDENT_COUNTER.join(prefixList) if indentionDeep == 1 \
        else INDENT_COUNTER.join(prefixList)

    isNew = True
    if str(type(object))[1] == 'c' or isinstance(object, dict):
        counter = 0
        for key, value in obj_dict.items():
            counter += 1
            # write the variable
            isNew = writeRow(key, value, prefixToWrite)

            # managing the extremes of the loop
            if counter == 1:
                isFirstOrLast = IS_FIRST
            elif counter == len(obj_dict):
                isFirstOrLast = IS_LAST
            else:
                isFirstOrLast = IS_MIDDLE

            # show attributes for objects and items for lists and tuples
            if isBelowMaxDeep and isNew and isIterable(value):
                recursivePrint(value, prefix, isFirstOrLast)
    else:
        for i in range(0, len(object)):
            # write the variable
            isNew = writeRow(object[i], object[i], prefixToWrite,
                             str(i + 1) + '/' + str(len(object)))

            # managing the extremes of the loop
            if i == 0:
                isFirstOrLast = IS_FIRST
            elif len(object) == i + 1:
                isFirstOrLast = IS_LAST
            else:
                isFirstOrLast = IS_MIDDLE

            # show attributes for objects and items for lists and tuples
            if isBelowMaxDeep and isNew and isIterable(object[i]):
                recursivePrint(object[i], prefix, isFirstOrLast)




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Tooltip widget, taken from web.
"""
        
import tkinter as tk

'''Michael Lange <klappnase (at) freakmail (dot) de>
The ToolTip class provides a flexible tooltip widget for Tkinter; it is based on IDLE's ToolTip
module which unfortunately seems to be broken (at least the version I saw).
INITIALIZATION OPTIONS:
anchor :        where the text should be positioned inside the widget, must be on of "n", "s", "e", "w", "nw" and so on;
                default is "center"
bd :            borderwidth of the widget; default is 1 (NOTE: don't use "borderwidth" here)
bg :            background color to use for the widget; default is "lightyellow" (NOTE: don't use "background")
delay :         time in ms that it takes for the widget to appear on the screen when the mouse pointer has
                entered the parent widget; default is 1500
fg :            foreground (i.e. text) color to use; default is "black" (NOTE: don't use "foreground")
follow_mouse :  if set to 1 the tooltip will follow the mouse pointer instead of being displayed
                outside of the parent widget; this may be useful if you want to use tooltips for
                large widgets like listboxes or canvases; default is 0
font :          font to use for the widget; default is system specific
justify :       how multiple lines of text will be aligned, must be "left", "right" or "center"; default is "left"
padx :          extra space added to the left and right within the widget; default is 4
pady :          extra space above and below the text; default is 2
relief :        one of "flat", "ridge", "groove", "raised", "sunken" or "solid"; default is "solid"
state :         must be "normal" or "disabled"; if set to "disabled" the tooltip will not appear; default is "normal"
text :          the text that is displayed inside the widget
textvariable :  if set to an instance of Tkinter.StringVar() the variable's value will be used as text for the widget
width :         width of the widget; the default is 0, which means that "wraplength" will be used to limit the widgets width
wraplength :    limits the number of characters in each line; default is 150

WIDGET METHODS:
configure(**opts) : change one or more of the widget's options as described above; the changes will take effect the
                    next time the tooltip shows up; NOTE: follow_mouse cannot be changed after widget initialization

Other widget methods that might be useful if you want to subclass ToolTip:
enter() :           callback when the mouse pointer enters the parent widget
leave() :           called when the mouse pointer leaves the parent widget
motion() :          is called when the mouse pointer moves inside the parent widget if follow_mouse is set to 1 and the
                    tooltip has shown up to continually update the coordinates of the tooltip window
coords() :          calculates the screen coordinates of the tooltip window
create_contents() : creates the contents of the tooltip window (by default a Tkinter.Label)
'''
# Ideas gleaned from PySol


class ToolTip:
    def __init__(self, master, text='Your text here', delay=1500, **opts):
        self.master = master
        self._opts = {'anchor': 'center', 'bd': 1, 'bg': 'lightyellow', 'delay': delay, 'fg': 'black',
                      'follow_mouse': 0, 'font': None, 'justify': 'left', 'padx': 4, 'pady': 2,
                      'relief': 'solid', 'state': 'normal', 'text': text, 'textvariable': None,
                      'width': 0, 'wraplength': 150}
        self.configure(**opts)
        self._tipwindow = None
        self._id = None
        self._id1 = self.master.bind("<Enter>", self.enter, '+')
        self._id2 = self.master.bind("<Leave>", self.leave, '+')
        self._id3 = self.master.bind("<ButtonPress>", self.leave, '+')
        self._follow_mouse = 0
        if self._opts['follow_mouse']:
            self._id4 = self.master.bind("<Motion>", self.motion, '+')
            self._follow_mouse = 1
    
    def configure(self, **opts):
        for key in opts:
            if key in self._opts:
                self._opts[key] = opts[key]
            else:    
                KeyError = 'KeyError: Unknown option: "%s"' % key
                raise KeyError
    
    # ----these methods handle the callbacks on "<Enter>", "<Leave>" and "<Motion>"-------------
    # ----events on the parent widget; override them if you want to change the widget's behavior
    
    def enter(self, event=None):
        self._schedule()
        
    def leave(self, event=None):
        self._unschedule()
        self._hide()
    
    def motion(self, event=None):
        if self._tipwindow and self._follow_mouse:
            x, y = self.coords()
            self._tipwindow.wm_geometry("+%d+%d" % (x, y))
    
    # ------the methods that do the work:---------------------------------------------------------##
    
    def _schedule(self):
        self._unschedule()
        if self._opts['state'] == 'disabled':
            return
        self._id = self.master.after(self._opts['delay'], self._show)

    def _unschedule(self):
        id = self._id
        self._id = None
        if id:
            self.master.after_cancel(id)

    def _show(self):
        if self._opts['state'] == 'disabled':
            self._unschedule()
            return
        if not self._tipwindow:
            self._tipwindow = tw = tk.Toplevel(self.master)
            # hide the window until we know the geometry
            tw.withdraw()
            tw.wm_overrideredirect(1)

            if tw.tk.call("tk", "windowingsystem") == 'aqua':
                tw.tk.call("::tk::unsupported::MacWindowStyle", "style", tw._w, "help", "none")

            self.create_contents()
            tw.update_idletasks()
            x, y = self.coords()
            tw.wm_geometry("+%d+%d" % (x, y))
            tw.deiconify()
    
    def _hide(self):
        tw = self._tipwindow
        self._tipwindow = None
        if tw:
            tw.destroy()
                
    ##----these methods might be overridden in derived classes:----------------------------------##
    
    def coords(self):
        # The tip window must be completely outside the master widget;
        # otherwise when the mouse enters the tip window we get
        # a leave event and it disappears, and then we get an enter
        # event and it reappears, and so on forever :-(
        # or we take care that the mouse pointer is always outside the tipwindow :-)
        tw = self._tipwindow
        twx, twy = tw.winfo_reqwidth(), tw.winfo_reqheight()
        w, h = tw.winfo_screenwidth(), tw.winfo_screenheight()
        # calculate the y coordinate:
        if self._follow_mouse:
            y = tw.winfo_pointery() + 20
            # make sure the tipwindow is never outside the screen:
            if y + twy > h:
                y = y - twy - 30
        else:
            y = self.master.winfo_rooty() + self.master.winfo_height() + 3
            if y + twy > h:
                y = self.master.winfo_rooty() - twy - 3
        # we can use the same x coord in both cases:
        x = tw.winfo_pointerx() - twx / 2
        if x < 0:
            x = 0
        elif x + twx > w:
            x = w - twx
        return x, y

    def create_contents(self):
        opts = self._opts.copy()
        for opt in ('delay', 'follow_mouse', 'state'):
            del opts[opt]
        label = tk.Label(self._tipwindow, **opts)
        label.pack()




#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import datetime as dt
import json
import os
import re
import time
import traceback
from collections import OrderedDict

import pyworkflow as pw
from pyworkflow.constants import PROJECT_DBNAME, PROJECT_SETTINGS
import pyworkflow.object as pwobj
import pyworkflow.protocol as pwprot
import pyworkflow.utils as pwutils
from pyworkflow.mapper import SqliteMapper
from pyworkflow.protocol.constants import (MODE_RESTART, MODE_RESUME,
                                           STATUS_INTERACTIVE, ACTIVE_STATUS,
                                           UNKNOWN_JOBID, INITIAL_SLEEP_TIME)
from pyworkflow.protocol.protocol import ProtImportBase

from . import config


OBJECT_PARENT_ID = pwobj.OBJECT_PARENT_ID
PROJECT_LOGS = 'Logs'
PROJECT_RUNS = 'Runs'
PROJECT_TMP = 'Tmp'
PROJECT_UPLOAD = 'Uploads'
PROJECT_CONFIG = '.config'
PROJECT_CREATION_TIME = 'CreationTime'

# Regex to get numbering suffix and automatically propose runName
REGEX_NUMBER_ENDING = re.compile('(?P<prefix>.+)(?P<number>\(\d*\))\s*$')
REGEX_NUMBER_ENDING_CP = re.compile('(?P<prefix>.+\s\(copy)(?P<number>.*)\)\s*$')


class Project(object):
    """This class will handle all information 
    related with a Project"""

    @classmethod
    def getDbName(cls):
        """ Return the name of the database file of projects. """
        return PROJECT_DBNAME

    def __init__(self, domain, path):
        """
        Create a new Project instance.
        :param domain: The application domain from where to get objects and
            protocols.
        :param path: Path where the project will be created/loaded
        """
        self._domain = domain
        self.name = path
        self.shortName = os.path.basename(path)
        self.path = os.path.abspath(path)
        self._isLink = os.path.islink(path)
        self._isInReadOnlyFolder = False
        self.pathList = []  # Store all related paths
        self.dbPath = self.__addPath(PROJECT_DBNAME)
        self.logsPath = self.__addPath(PROJECT_LOGS)
        self.runsPath = self.__addPath(PROJECT_RUNS)
        self.tmpPath = self.__addPath(PROJECT_TMP)
        self.uploadPath = self.__addPath(PROJECT_UPLOAD)
        self.settingsPath = self.__addPath(PROJECT_SETTINGS)
        self.configPath = self.__addPath(PROJECT_CONFIG)
        self.runs = None
        self._runsGraph = None
        self._transformGraph = None
        self._sourceGraph = None
        self.address = ''
        self.port = pwutils.getFreePort()
        self.mapper = None
        self.settings = None
        # Host configuration
        self._hosts = None
        #  Creation time should be stored in project.sqlite when the project
        # is created and then loaded with other properties from the database
        self._creationTime = None
        # Time stamp with the last run has been updated
        self._lastRunTime = None

    def getObjId(self):
        """ Return the unique id assigned to this project. """
        return os.path.basename(self.path)

    def __addPath(self, *paths):
        """Store a path needed for the project"""
        p = self.getPath(*paths)
        self.pathList.append(p)
        return p

    def getPath(self, *paths):
        """Return path from the project root"""
        if paths:
            return os.path.join(*paths)
        else:
            return self.path

    def isLink(self):
        """Returns if the project path is a link to another folder."""
        return self._isLink

    def getDbPath(self):
        """ Return the path to the sqlite db. """
        return self.dbPath

    def getDbLastModificationDate(self):
        """ Return the last modification date of the database """
        pwutils.getFileLastModificationDate(self.getDbPath())

    def getCreationTime(self):
        """ Return the time when the project was created. """
        # In project.create method, the first object inserted
        # in the mapper should be the creation time
        return self._creationTime

    def getSettingsCreationTime(self):
        return self.settings.getCreationTime()

    def getElapsedTime(self):
        """ Returns the time elapsed from the creation to the last
        execution time. """
        if self._creationTime and self._lastRunTime:
            creationTs = self._creationTime
            lastRunTs = self._lastRunTime.datetime()
            return lastRunTs - creationTs
        return None

    def getLeftTime(self):
        lifeTime = self.settings.getLifeTime()
        if lifeTime:
            td = dt.timedelta(hours=lifeTime)
            return td - self.getElapsedTime()
        else:
            return None

    def setDbPath(self, dbPath):
        """ Set the project db path.
        This function is used when running a protocol where
        a project is loaded but using the protocol own sqlite file.
        """
        # First remove from pathList the old dbPath
        self.pathList.remove(self.dbPath)
        self.dbPath = os.path.abspath(dbPath)
        self.pathList.append(self.dbPath)

    def getName(self):
        return self.name

    def getDomain(self):
        return self._domain

    # TODO: maybe it has more sense to use this behaviour
    # for just getName function...
    def getShortName(self):
        return self.shortName

    def getTmpPath(self, *paths):
        return self.getPath(PROJECT_TMP, *paths)

    def getLogPath(self, *paths):
        return self.getPath(PROJECT_LOGS, *paths)

    def getSettings(self):
        return self.settings

    def saveSettings(self):
        # Read only mode
        if not self.openedAsReadOnly():
            self.settings.write()

    def createSettings(self, runsView=1, readOnly=False):
        self.settings = config.ProjectSettings()
        self.settings.setRunsView(runsView)
        self.settings.setReadOnly(readOnly)
        self.settings.write(self.settingsPath)
        return self.settings

    def createMapper(self, sqliteFn):
        """ Create a new SqliteMapper object and pass as classes dict
        all globals and update with data and protocols from em.
        """
        classesDict = pwobj.Dict(default=pwprot.LegacyProtocol)
        classesDict.update(self._domain.getMapperDict())
        classesDict.update(config.__dict__)
        return SqliteMapper(sqliteFn, classesDict)

    def load(self, dbPath=None, hostsConf=None, protocolsConf=None, chdir=True,
             loadAllConfig=True):
        """ Load project data, configuration and settings.
        Params:
            dbPath: the path to the project database.
                If None, use the project.sqlite in the project folder.
            hosts: where to read the host configuration. 
                If None, check if exists in .config/hosts.conf
                or read from ~/.config/scipion/hosts.conf
            settings: where to read the settings.
                If None, use the settings.sqlite in project folder.
                If forProtocol is True, the settings and protocols.conf will
                not be loaded.
        """

        if not os.path.exists(self.path):
            raise Exception("Cannot load project, path doesn't exist: %s"
                            % self.path)

        # If folder is read only, flag it and warn about it.
        if not os.access(self.path, os.W_OK):
            self._isInReadOnlyFolder = True
            print("WARNING on project \"%s\": don't have write permissions "
                  "for project folder. Loading as READ-ONLY." % self.shortName)

        if chdir:
            os.chdir(self.path)  # Before doing nothing go to project dir

        try:
            self._loadDb(dbPath)
            self._loadHosts(hostsConf)

            if loadAllConfig:

                # FIXME: Handle settings argument here

                # It is possible that settings does not exists if
                # we are loading a project after a Project.setDbName,
                # used when running protocols
                settingsPath = os.path.join(self.path, self.settingsPath)
                if pw.Config.debugOn():
                    print("settingsPath: %s" % settingsPath)

                if os.path.exists(settingsPath):
                    self.settings = config.ProjectSettings.load(settingsPath)
                else:
                    print("settings is None")
                    self.settings = None

            self._loadCreationTime()

        # Catch DB not found exception (when loading a project from a folder
        #  without project.sqlite
        except MissingProjectDbException as noDBe:
            # Raise it at before: This is a critical error and should be raised
            raise noDBe

        # Catch any less severe exception..to allow at least open the project.
        # except Exception as e:
        #     print("ERROR: Project %s load failed.\n"
        #           "       Message: %s\n" % (self.path, e))

    def _loadCreationTime(self):
        # Load creation time, it should be in project.sqlite or
        # in some old projects it is found in settings.sqlite

        creationTime = self.mapper.selectBy(name=PROJECT_CREATION_TIME)

        if creationTime:  # CreationTime was found in project.sqlite
            self._creationTime = creationTime[0].datetime()
        else:
            # We should read the creation time from settings.sqlite and
            # update the CreationTime in the project.sqlite
            self._creationTime = self.getSettingsCreationTime()
            self._storeCreationTime(self._creationTime)

    # ---- Helper functions to load different pieces of a project
    def _loadDb(self, dbPath):
        """ Load the mapper from the sqlite file in dbPath. """
        if dbPath is not None:
            self.setDbPath(dbPath)

        absDbPath = os.path.join(self.path, self.dbPath)
        if not os.path.exists(absDbPath):
            raise MissingProjectDbException(
                "Project database not found at '%s'" % absDbPath)
        self.mapper = self.createMapper(absDbPath)

    def closeMapper(self):
        if self.mapper is not None:
            self.mapper.close()
            self.mapper = None

    def getLocalConfigHosts(self):
        """ Return the local file where the project will try to
        read the hosts configuration. """
        return self.getPath(PROJECT_CONFIG, pw.Config.SCIPION_HOSTS)

    def _loadHosts(self, hosts):
        """ Loads hosts configuration from hosts file. """
        # If the host file is not passed as argument...
        configHosts = pw.Config.SCIPION_HOSTS
        projHosts = self.getLocalConfigHosts()

        if hosts is None:
            # Try first to read it from the project file .config./hosts.conf
            if os.path.exists(projHosts):
                hostsFile = projHosts
            else:
                localDir = os.path.dirname(pw.Config.SCIPION_LOCAL_CONFIG)
                hostsFile = os.path.join(localDir, configHosts)
        else:
            pwutils.copyFile(hosts, projHosts)
            hostsFile = hosts

        self._hosts = pwprot.HostConfig.load(hostsFile)

    def getHostNames(self):
        """ Return the list of host name in the project. """
        return list(self._hosts.keys())

    def getHostConfig(self, hostName):
        if hostName in self._hosts:
            hostKey = hostName
        else:
            hostKey = self.getHostNames()[0]
            print("PROJECT: Warning, protocol host '%s' not found." % hostName)
            print("         Using '%s' instead." % hostKey)

        return self._hosts[hostKey]

    def getProtocolView(self):
        """ Returns de view selected in the tree when it was persisted"""
        return self.settings.getProtocolView()

    def create(self, runsView=1, readOnly=False, hostsConf=None,
               protocolsConf=None):
        """Prepare all required paths and files to create a new project.
        Params:
         hosts: a list of configuration hosts associated to this projects
               (class ExecutionHostConfig)
        """
        # Create project path if not exists
        pwutils.path.makePath(self.path)
        os.chdir(self.path)  # Before doing nothing go to project dir
        self._cleanData()
        print("Creating project at: ", os.path.abspath(self.dbPath))
        # Create db through the mapper
        self.mapper = self.createMapper(self.dbPath)
        # Store creation time
        self._storeCreationTime(dt.datetime.now())
        # Load settings from .conf files and write .sqlite
        self.settings = self.createSettings(runsView=runsView,
                                            readOnly=readOnly)
        # Create other paths inside project
        for p in self.pathList:
            pwutils.path.makePath(p)

        self._loadHosts(hostsConf)

    def _storeCreationTime(self, creationTime):
        """ Store the creation time in the project db. """
        # Store creation time
        creation = pwobj.String(objName=PROJECT_CREATION_TIME)
        creation.set(creationTime)
        self.mapper.insert(creation)
        self.mapper.commit()

    def _cleanData(self):
        """Clean all project data"""
        pwutils.path.cleanPath(*self.pathList)

    def _continueWorkflow(self, continuedProtList=None, errorsList=None):
        """
        This function continue a workflow from a selected protocol.
        The previous results are preserved.
        Actions done here are:
        1. if the protocol list exists (for each protocol)
            1.1  if the protocol is not an interactive protocol
            1.1.1. If the protocol is in streaming (CONTINUE ACTION):
                       - 'dataStreaming' parameter if the protocol is an import
                          protocol
                       -  check if the __stepsCheck function exist and it's not
                          the same implementation of the base class
                          (worksInStreaming function)
                        1.1.1.1 Open the protocol sets, store and save them in
                                the  database
                       1.1.1.2 Change the protocol status (SAVED)
                       1.1.1.3 Schedule the protocol
                   Else Restart the workflow from that point (RESTART ACTION) if
                   at least one protocol in streaming has been launched
        """
        if continuedProtList is not None:
            for protocol, level in continuedProtList.values():
                if not protocol.isInteractive():
                    if protocol.worksInStreaming():
                        attrSet = [attr for name, attr in
                                   protocol.iterOutputAttributes(pwprot.Set)]
                        try:
                            if attrSet:
                                for attr in attrSet:
                                    attr.setStreamState(attr.STREAM_OPEN)
                                    attr.write()
                                    attr.close()
                            protocol.setStatus(pwprot.STATUS_SAVED)
                            protocol._updateSteps(lambda step: step.setStatus(pwprot.STATUS_SAVED))
                            protocol.setMapper(self.createMapper(protocol.getDbPath()))
                            protocol._store()
                            self._storeProtocol(protocol)
                            self.scheduleProtocol(protocol,
                                                  initialSleepTime=level*INITIAL_SLEEP_TIME)
                        except Exception as ex:
                            errorsList.append("Error trying to launch the "
                                              "protocol: %s\nERROR: %s\n" %
                                              (protocol.getObjLabel(), ex))
                            break
                    else:
                        if level != 0:
                            # we make sure that at least one protocol in streaming
                            # has been launched
                            self._restartWorkflow({protocol.getObjId(): (protocol, level)},
                                                  errorsList)

                        else:
                            errorsList.append(("Error trying to launch the "
                                               "protocol: %s\nERROR: The protocol is "
                                               "not in streaming" %
                                               (protocol.getObjLabel())))
                            break

    def _restartWorkflow(self, restartedProtList=None, errorsList=None):
        """
        This function restart a workflow from a selected protocol.
        All previous results will be deleted
        Actions done here are:
        1. Set the protocol run mode (RESTART). All previous results will be
           deleted
        2. Schedule the protocol if not is an interactive protocol
        3. For each of the dependents protocols, repeat from step 1
        """
        if restartedProtList is not None:
            for protocol, level in restartedProtList.values():
                if not protocol.isInteractive():
                    try:
                        protocol.runMode.set(MODE_RESTART)
                        self.scheduleProtocol(protocol,
                                              initialSleepTime=level*INITIAL_SLEEP_TIME)
                    except Exception as ex:
                        errorsList.append("Error trying to restart a protocol: %s"
                                          "\nERROR: %s\n" % (protocol.getObjLabel(),
                                                             ex))
                        break
                else:
                    protocol.setStatus(pwprot.STATUS_SAVED)
                    self._storeProtocol(protocol)
                    protocol.runMode.set(MODE_RESTART)
                    self._setupProtocol(protocol)
                    protocol.makePathsAndClean()  # Create working dir if necessary
                    # Delete the relations created by this protocol
                    self.mapper.deleteRelations(self)
                    self.mapper.commit()
                    self.mapper.store(protocol)
                    self.mapper.commit()

    def _fixProtParamsConfiguration(self, protocol=None):
        """
        This function fix:
        1. The old parameters configuration in the protocols.
           Now, dependent protocols have a pointer to the parent protocol, and
           the extended parameter has a parent output value
        """
        # Take the old configuration attributes and fix the pointer
        oldStylePointerList = [item for key, item in
                               protocol.iterInputAttributes()
                               if not isinstance(item.getObjValue(),
                                                 pwprot.Protocol)]
        if oldStylePointerList:
            # Fix the protocol parameters
            for pointer in oldStylePointerList:
                auxPointer = pointer.getObjValue()
                pointer.set(self.getRunsGraph().getNode(str(pointer.get().getObjParentId())).run)
                pointer.setExtended(auxPointer.getLastName())
                protocol._store()
                self._storeProtocol(protocol)
                self._updateProtocol(protocol)
                self.mapper.commit()

    def stopWorkFlow(self, activeProtList):
        """
        This function can stop a workflow from a selected protocol
        :param initialProtocol: selected protocol
        """
        errorProtList = []
        for protocol in activeProtList:
            try:
                self.stopProtocol(protocol)
            except Exception:
                errorProtList.append(protocol)
        return errorProtList

    def resetWorkFlow(self, workflowProtocolList):
        """
        This function can reset a workflow from a selected protocol
        :param initialProtocol: selected protocol
        """
        errorProtList = []
        if workflowProtocolList:
            for protocol, level in workflowProtocolList.values():
                if protocol.getStatus() != pwprot.STATUS_SAVED:
                    try:
                        self.resetProtocol(protocol)
                    except Exception:
                        errorProtList.append(protocol)
        return errorProtList

    def launchWorkflow(self, workflowProtocolList, mode=MODE_RESUME):
        """
        This function can launch a workflow from a selected protocol in two
        modes depending on the 'mode' value (RESTART, CONTINUE)
        Actions done here are:
        1. Check if the workflow has active protocols.
        2. Fix the workflow if is not properly configured
        3. Restart or Continue a workflow starting from the protocol depending
           of the 'mode' value
        """
        errorsList = []
        if mode == MODE_RESTART:
            self._restartWorkflow(workflowProtocolList, errorsList)
        else:
            self._continueWorkflow(workflowProtocolList, errorsList)
        return errorsList

    def launchProtocol(self, protocol, wait=False, scheduled=False,
                       force=False):
        """ In this function the action of launching a protocol
        will be initiated. Actions done here are:
        1. Store the protocol and assign name and working dir
        2. Create the working dir and also the protocol independent db
        3. Call the launch method in protocol.job to handle submission:
           mpi, thread, queue,
        and also take care if the execution is remotely.

        If the protocol has some prerequisites (other protocols that
        needs to be finished first), it will be scheduled.
        """
        if protocol.getPrerequisites() and not scheduled:
            return self.scheduleProtocol(protocol)

        isRestart = protocol.getRunMode() == MODE_RESTART

        if not force:
            if (not protocol.isInteractive() and not protocol.isInStreaming()) or isRestart:
                self._checkModificationAllowed([protocol],
                                               'Cannot RE-LAUNCH protocol')

        protocol.setStatus(pwprot.STATUS_LAUNCHED)
        self._setupProtocol(protocol)

        # Prepare a separate db for this run if not from schedule jobs
        # Scheduled protocols will load the project db from the run.db file,
        # so there is no need to copy the database

        if not scheduled:
            protocol.makePathsAndClean()  # Create working dir if necessary
            # Delete the relations created by this protocol
            if isRestart:
                self.mapper.deleteRelations(self)
            self.mapper.commit()

            # NOTE: now we are simply copying the entire project db, this can be
            # changed later to only create a subset of the db need for the run
            pwutils.path.copyFile(self.dbPath, protocol.getDbPath())

        # Launch the protocol, the jobId should be set after this call
        jobId = pwprot.launch(protocol, wait)
        if jobId is None or jobId == UNKNOWN_JOBID:
            protocol.setStatus(pwprot.STATUS_FAILED)

        # Commit changes
        if wait:  # This is only useful for launching tests...
            self._updateProtocol(protocol)
        else:
            self.mapper.store(protocol)
        self.mapper.commit()

    def scheduleProtocol(self, protocol, prerequisites=[], initialSleepTime=0):
        """ Schedule a new protocol that will run when the input data
        is available and the prerequisites are finished.
        Params:
            protocol: the protocol that will be scheduled.
            prerequisites: a list with protocols ids that the scheduled
                protocol will wait for.
        """
        isRestart = protocol.getRunMode() == MODE_RESTART

        protocol.setStatus(pwprot.STATUS_SCHEDULED)
        protocol.addPrerequisites(*prerequisites)

        self._setupProtocol(protocol)
        protocol.makePathsAndClean()  # Create working dir if necessary
        # Delete the relations created by this protocol if any
        if isRestart:
            self.mapper.deleteRelations(self)
        self.mapper.commit()

        # Prepare a separate db for this run
        # NOTE: now we are simply copying the entire project db, this can be
        # changed later to only create a subset of the db need for the run
        pwutils.path.copyFile(self.dbPath, protocol.getDbPath())
        # Launch the protocol, the jobId should be set after this call
        pwprot.schedule(protocol, initialSleepTime=initialSleepTime)
        self.mapper.store(protocol)
        self.mapper.commit()

    def _updateProtocol(self, protocol, tries=0, checkPid=False,
                        skipUpdatedProtocols=True):

        # If this is read only exit
        if self.openedAsReadOnly():
            return pw.NOT_UPDATED_READ_ONLY

        if checkPid:
            self.checkPid(protocol)

        if skipUpdatedProtocols:
            # If we are already updated, comparing timestamps
            if pwprot.isProtocolUpToDate(protocol):
                return pw.NOT_UPDATED_UNNECESSARY

        try:
            # Backup the values of 'jobId', 'label' and 'comment'
            # to be restored after the .copy
            jobId = protocol.getJobId()
            label = protocol.getObjLabel()
            comment = protocol.getObjComment()

            # If the protocol database has ....
            #  Comparing date will not work unless we have a reliable
            # lastModificationDate of a protocol in the project.sqlite
            # TODO: when launching remote protocols, the db should be
            # TODO: retrieved in a different way.
            prot2 = pwprot.getProtocolFromDb(self.path,
                                             protocol.getDbPath(),
                                             protocol.getObjId())

            # Capture the db timestamp before loading.
            lastUpdateTime = pwutils.getFileLastModificationDate(protocol.getDbPath())

            # Copy is only working for db restored objects
            protocol.setMapper(self.mapper)

            localOutputs = list(protocol._outputs)
            protocol.copy(prot2, copyId=False, excludeInputs=True)

            # merge outputs: This is necessary when outputs are added from the GUI
            # e.g.: adding coordinates from analyze result and protocol is active (interactive).
            for attr in localOutputs:
                if attr not in protocol._outputs:
                    protocol._outputs.append(attr)

            # Restore backup values
            protocol.setJobId(jobId)
            protocol.setObjLabel(label)
            protocol.setObjComment(comment)
            # Use the run.db timestamp instead of the system TS to prevent
            # possible inconsistencies
            # protocol.lastUpdateTimeStamp.set(datetime.datetime.now())
            protocol.lastUpdateTimeStamp.set(lastUpdateTime)

            self.mapper.store(protocol)

            # Close DB connections
            prot2.getProject().closeMapper()
            prot2.closeMappers()

        except Exception as ex:
            print("Error trying to update protocol: %s(jobId=%s)\n "
                  "ERROR: %s, tries=%d"
                  % (protocol.getObjName(), jobId, ex, tries))
            if tries == 3:  # 3 tries have been failed
                traceback.print_exc()
                # If any problem happens, the protocol will be marked
                # with a FAILED status
                protocol.setFailed(str(ex))
                self.mapper.store(protocol)
                return pw.NOT_UPDATED_ERROR
            else:
                time.sleep(0.5)
                self._updateProtocol(protocol, tries + 1)

        return pw.PROTOCOL_UPDATED

    def stopProtocol(self, protocol):
        """ Stop a running protocol """
        try:
            if protocol.getStatus() in ACTIVE_STATUS:
                pwprot.stop(protocol)
        except Exception as e:
            print("stopProtocol: error", e)
            raise
        finally:
            protocol.setAborted()
            protocol.setMapper(self.createMapper(protocol.getDbPath()))
            protocol._store()
            self._storeProtocol(protocol)
            protocol.getMapper().close()

    def resetProtocol(self, protocol):
        """ Stop a running protocol """
        try:
            if protocol.getStatus() in ACTIVE_STATUS:
                pwprot.stop(protocol)
        except Exception:
            raise
        finally:
            protocol.setSaved()
            protocol.runMode.set(MODE_RESTART)
            protocol._store()
            self._storeProtocol(protocol)
            protocol.makePathsAndClean()  # Create working dir if necessary
            protocol._store()
            self._storeProtocol(protocol)

    def continueProtocol(self, protocol):
        """ This function should be called 
        to mark a protocol that have an interactive step
        waiting for approval that can continue
        """
        protocol.continueFromInteractive()
        self.launchProtocol(protocol)

    def __protocolInList(self, prot, protocols):
        """ Check if a protocol is in a list comparing the ids. """
        for p in protocols:
            if p.getObjId() == prot.getObjId():
                return True
        return False

    def __validDependency(self, prot, child, protocols):
        """ Check if the given child is a true dependency of the protocol
        in order to avoid any modification.
        """
        return (not self.__protocolInList(child, protocols) and
                not child.isSaved() and not child.isScheduled())

    def _getProtocolsDependencies(self, protocols):
        error = ''
        runsGraph = self.getRunsGraph()
        for prot in protocols:
            node = runsGraph.getNode(prot.strId())
            if node:
                childs = [node.run for node in node.getChilds() if
                          self.__validDependency(prot, node.run, protocols)]
                if childs:
                    deps = [' ' + c.getRunName() for c in childs]
                    error += '\n *%s* is referenced from:\n   - ' % prot.getRunName()
                    error += '\n   - '.join(deps)
        return error

    def _checkProtocolsDependencies(self, protocols, msg):
        """ Check if the protocols have dependencies.
        This method is used before delete or save protocols to be sure
        it is not referenced from other runs. (an Exception is raised)
        Params:
             protocols: protocol list to be analyzed.
             msg: String message to be prefixed to Exception error.
        """
        # Check if the protocol have any dependencies
        error = self._getProtocolsDependencies(protocols)
        if error:
            raise ModificationNotAllowedException(msg + error)

    def _checkModificationAllowed(self, protocols, msg):
        """ Check if any modification operation is allowed for
        this group of protocols. 
        """
        if self.openedAsReadOnly():
            raise Exception(msg + " Running in READ-ONLY mode.")

        self._checkProtocolsDependencies(protocols, msg)

    def _getWorkflowFromProtocol(self, protocol, fixProtParam=True):
        """
        This function get the workflow from "protocol" and determine the
        protocol level into the graph. Also, checks if there are active
        protocols excluding interactive protocols.
        """
        activeProtList = []
        configuredProtList = {}
        auxProtList = []
        # store the protocol and your level into the workflow
        configuredProtList[protocol.getObjId()] = [protocol, 0]
        auxProtList.append(protocol.getObjId())
        runGraph = self.getRunsGraph()

        while auxProtList:
            protocol = runGraph.getNode(str(auxProtList.pop(0))).run
            level = configuredProtList[protocol.getObjId()][1] + 1
            if fixProtParam:
                self._fixProtParamsConfiguration(protocol)
            if protocol.isActive() and protocol.getStatus() != STATUS_INTERACTIVE:
                activeProtList.append(protocol)
            node = runGraph.getNode(protocol.strId())
            dependencies = [node.run for node in node.getChilds()]
            for dep in dependencies:
                if not dep.getObjId() in auxProtList:
                    auxProtList.append(dep.getObjId())
                if not dep.getObjId() in configuredProtList.keys():
                    configuredProtList[dep.getObjId()] = [dep, level]
                elif level > configuredProtList[dep.getObjId()][1]:
                    configuredProtList[dep.getObjId()][1] = level

        return configuredProtList, activeProtList

    def deleteProtocol(self, *protocols):
        self._checkModificationAllowed(protocols, 'Cannot DELETE protocols')

        for prot in protocols:
            # Delete the relations created by this protocol
            self.mapper.deleteRelations(prot)
            # Delete from protocol from database
            self.mapper.delete(prot)
            wd = prot.workingDir.get()

            if wd.startswith(PROJECT_RUNS):
                prot.cleanWorkingDir()
            else:
                print("Error path: ", wd)

        self.mapper.commit()

    def deleteProtocolOutput(self, protocol, output):
        """ Delete a given object from the project.
        Usually to clean up some outputs.
        """
        node = self.getRunsGraph().getNode(protocol.strId())
        deps = []

        for node in node.getChilds():
            for _, inputObj in node.run.iterInputAttributes():
                value = inputObj.get()
                if (value is not None and
                        value.getObjId() == output.getObjId() and
                        not node.run.isSaved()):
                    deps.append(node.run)

        if deps:
            error = 'Cannot DELETE Object, it is referenced from:'
            for d in deps:
                error += '\n - %s' % d.getRunName()
            raise Exception(error)
        else:
            protocol.deleteOutput(output)
            pwutils.path.copyFile(self.dbPath, protocol.getDbPath())

    def __setProtocolLabel(self, newProt):
        """ Set a readable label to a newly created protocol.
        We will try to find another existing protocol with the default label
        and then use an incremental labeling in parenthesis (<number>++)
        """
        defaultLabel = newProt.getClassLabel()
        maxSuffix = 0

        for prot in self.getRuns(iterate=True, refresh=False):
            otherProtLabel = prot.getObjLabel()
            m = REGEX_NUMBER_ENDING.match(otherProtLabel)
            if m and m.groupdict()['prefix'].strip() == defaultLabel:
                stringSuffix = m.groupdict()['number'].strip('(').strip(')')
                try:
                    maxSuffix = max(int(stringSuffix), maxSuffix)
                except:
                    print("stringSuffix", stringSuffix)
            elif otherProtLabel == defaultLabel:  # When only we have the prefix,
                maxSuffix = max(1, maxSuffix)     # this REGEX don't match.

        if maxSuffix:
            protLabel = '%s (%d)' % (defaultLabel, maxSuffix+1)
        else:
            protLabel = defaultLabel

        newProt.setObjLabel(protLabel)

    def newProtocol(self, protocolClass, **kwargs):
        """ Create a new protocol from a given class. """
        newProt = protocolClass(project=self, **kwargs)
        # Only set a default label to the protocol if is was not
        # set through the kwargs
        if not newProt.getObjLabel():
            self.__setProtocolLabel(newProt)

        newProt.setMapper(self.mapper)
        newProt.setProject(self)

        return newProt

    def __getIOMatches(self, node, childNode):
        """ Check if some output of node is used as input in childNode.
        Return the list of attribute names that matches.
        Used from self.copyProtocol
        """
        matches = []
        for iKey, iAttr in childNode.run.iterInputAttributes():
            # As this point iAttr should be always a Pointer that 
            # points to the output of other protocol
            if iAttr.getObjValue() is node.run:
                oKey = iAttr.getExtended()
                matches.append((oKey, iKey))
            else:
                for oKey, oAttr in node.run.iterOutputAttributes():
                    # If node output is "real" and iAttr is still just a pointer
                    # the iAttr.get() will return None
                    pointed = iAttr.get()
                    if pointed is not None and oAttr.getObjId() == pointed.getObjId():
                        matches.append((oKey, iKey))

        return matches

    def __cloneProtocol(self, protocol):
        """ Make a copy of the protocol parameters, not outputs. 
            We will label the new protocol with the same name adding the 
            parenthesis as follow -> (copy) -> (copy 2) -> (copy 3)
        """
        newProt = self.newProtocol(protocol.getClass())
        oldProtName = protocol.getRunName()
        maxSuffix = 0

        # if '(copy...' suffix is not in the old name, we add it in the new name
        # and setting the newnumber
        mOld = REGEX_NUMBER_ENDING_CP.match(oldProtName)
        if mOld:
            newProtPrefix = mOld.groupdict()['prefix']
            if mOld.groupdict()['number'] == '':
                oldNumber = 1
            else:
                oldNumber = int(mOld.groupdict()['number'])
        else:
            newProtPrefix = oldProtName + ' (copy'
            oldNumber = 0
        newNumber = oldNumber + 1

        # looking for "<old name> (copy" prefixes in the project and
        # setting the newNumber as the maximum+1
        for prot in self.getRuns(iterate=True, refresh=False):
            otherProtLabel = prot.getObjLabel()
            mOther = REGEX_NUMBER_ENDING_CP.match(otherProtLabel)
            if mOther and mOther.groupdict()['prefix'] == newProtPrefix:
                stringSuffix = mOther.groupdict()['number']
                if stringSuffix == '':
                    stringSuffix = 1
                maxSuffix = max(maxSuffix, int(stringSuffix))
                if newNumber <= maxSuffix:
                    newNumber = maxSuffix + 1

        # building the new name
        if newNumber == 1:
            newProtLabel = newProtPrefix + ')'
        else:
            newProtLabel = '%s %d)' % (newProtPrefix, newNumber)

        newProt.setObjLabel(newProtLabel)
        newProt.copyDefinitionAttributes(protocol)
        newProt.copyAttributes(protocol, 'hostName', '_useQueue', '_queueParams')
        newProt.runMode.set(MODE_RESTART)

        return newProt

    def copyProtocol(self, protocol):
        """ Make a copy of the protocol,
        Return a new instance with copied values. """
        result = None

        if isinstance(protocol, pwprot.Protocol):
            result = self.__cloneProtocol(protocol)

        elif isinstance(protocol, list):
            # Handle the copy of a list of protocols
            # for this case we need to update the references of input/outputs
            newDict = {}

            for prot in protocol:
                newProt = self.__cloneProtocol(prot)
                newDict[prot.getObjId()] = newProt
                self.saveProtocol(newProt)

            g = self.getRunsGraph()

            for prot in protocol:
                node = g.getNode(prot.strId())
                newProt = newDict[prot.getObjId()]

                for childNode in node.getChilds():
                    newChildProt = newDict.get(childNode.run.getObjId(), None)

                    if newChildProt:
                        # Get the matches between outputs/inputs of
                        # node and childNode
                        matches = self.__getIOMatches(node, childNode)
                        # For each match, set the pointer and the extend
                        # attribute to reproduce the dependencies in the
                        # new workflow
                        for oKey, iKey in matches:
                            childPointer = getattr(newChildProt, iKey)

                            # Scalar with pointer case: If is a scalar with a pointer
                            if isinstance(childPointer, pwobj.Scalar) and childPointer.hasPointer():
                              # In this case childPointer becomes the contained Pointer
                              childPointer = childPointer.getPointer()

                            elif isinstance(childPointer, pwobj.PointerList):
                                for p in childPointer:
                                    if p.getObjValue().getObjId() == prot.getObjId():
                                        childPointer = p
                            childPointer.set(newProt)
                            childPointer.setExtended(oKey)
                        self.mapper.store(newChildProt)

            self.mapper.commit()
        else:
            raise Exception("Project.copyProtocol: invalid input protocol ' "
                            "'type '%s'." % type(protocol))

        return result

    def getProtocolsDict(self, protocols=None, namesOnly=False):
        """ Create a dict with the information of the given protocols.
         Params:
            protocols: list of protocols or None to include all.
            namesOnly: the output list will contain only the protocol names.
        """
        protocols = protocols or self.getRuns()

        # If the nameOnly, we will simply return a json list with their names
        if namesOnly:
            return {i: prot.getClassName() for i, prot in enumerate(protocols)}

        # Handle the copy of a list of protocols
        # for this case we need to update the references of input/outputs
        newDict = OrderedDict()

        for prot in protocols:
            newDict[prot.getObjId()] = prot.getDefinitionDict()

        g = self.getRunsGraph()

        # pwutils.startDebugger('a')
        for prot in protocols:
            protId = prot.getObjId()
            node = g.getNode(prot.strId())

            for childNode in node.getChilds():
                childId = childNode.run.getObjId()
                childProt = childNode.run
                if childId in newDict:
                    childDict = newDict[childId]
                    # Get the matches between outputs/inputs of
                    # node and childNode
                    matches = self.__getIOMatches(node, childNode)
                    for oKey, iKey in matches:
                        inputAttr = getattr(childProt, iKey)
                        if isinstance(inputAttr, pwobj.PointerList):
                            childDict[iKey] = [p.getUniqueId() for p in
                                               inputAttr]
                        else:
                            childDict[iKey] = '%s.%s' % (
                                protId, oKey)  # equivalent to pointer.getUniqueId

        return newDict

    def getProtocolsJson(self, protocols=None, namesOnly=False):
        """ Wraps getProtocolsDict to get a json string
             Params:
                protocols: list of protocols or None to include all.
                namesOnly: the output list will contain only the protocol names.
        """
        newDict = self.getProtocolsDict(protocols=protocols, namesOnly=namesOnly)
        return json.dumps(list(newDict.values()),
                          indent=4, separators=(',', ': '))

    def exportProtocols(self, protocols, filename):
        """ Create a text json file with the info
        to import the workflow into another project.
        This methods is very similar to copyProtocol
        Params:
            protocols: a list of protocols to export.
            filename: the filename where to write the workflow.
        """
        jsonStr = self.getProtocolsJson(protocols)
        f = open(filename, 'w')
        f.write(jsonStr)
        f.close()

    def loadProtocols(self, filename=None, jsonStr=None):
        """ Load protocols generated in the same format as self.exportProtocols.
        Params:
            filename: the path of the file where to read the workflow.
            jsonStr: read the protocols from a string instead of file.
        Note: either filename or jsonStr should be not None.
        """
        f = open(filename)
        importDir = os.path.dirname(filename)
        protocolsList = json.load(f)

        emProtocols = self._domain.getProtocols()
        newDict = OrderedDict()

        # First iteration: create all protocols and setup parameters
        for i, protDict in enumerate(protocolsList):
            protClassName = protDict['object.className']
            protId = protDict['object.id']
            protClass = emProtocols.get(protClassName, None)

            if protClass is None:
                print("ERROR: protocol class name '%s' not found" % protClassName)
            else:
                protLabel = protDict.get('object.label', None)
                prot = self.newProtocol(protClass,
                                        objLabel=protLabel,
                                        objComment=protDict.get('object.comment', None))
                protocolsList[i] = prot.processImportDict(protDict, importDir)

                prot._useQueue.set(protDict.get('_useQueue', False))
                prot._queueParams.set(protDict.get('_queueParams', None))
                prot._prerequisites.set(protDict.get('_prerequisites', None))
                prot.forceSchedule.set(protDict.get('forceSchedule', False))
                newDict[protId] = prot
                self.saveProtocol(prot)

        # Second iteration: update pointers values
        def _setPointer(pointer, value):
            # Properly setup the pointer value checking if the 
            # id is already present in the dictionary
            # Value to pointers could be None: Partial workflows
            if value:
                parts = value.split('.')
                target = newDict.get(parts[0], None)
                pointer.set(target)
                if not pointer.pointsNone():
                    pointer.setExtendedParts(parts[1:])

        def _setPrerequisites(prot):
            prerequisites = prot.getPrerequisites()
            if prerequisites:
                newPrerequisites = []
                for prerequisite in prerequisites:
                    if prerequisite in newDict:
                        newProtId = newDict[prerequisite].getObjId()
                        newPrerequisites.append(newProtId)
                    else:
                        print('Wait for id %s missing: ignored' % prerequisite)
                prot._prerequisites.set(newPrerequisites)

        for protDict in protocolsList:
            protId = protDict['object.id']

            if protId in newDict:
                prot = newDict[protId]
                _setPrerequisites(prot)
                for paramName, attr in prot.iterDefinitionAttributes():
                    if paramName in protDict:
                        # If the attribute is a pointer, we should look
                        # if the id is already in the dictionary and 
                        # set the extended property
                        if attr.isPointer():
                            _setPointer(attr, protDict[paramName])
                        # This case is similar to Pointer, but the values
                        # is a list and we will setup a pointer for each value
                        elif isinstance(attr, pwobj.PointerList):
                            attribute = protDict[paramName]
                            if attribute is None:
                                continue
                            for value in attribute:
                                p = pwobj.Pointer()
                                _setPointer(p, value)
                                attr.append(p)
                        # For "normal" parameters we just set the string value
                        else:
                            try:
                                attr.set(protDict[paramName])
                            # Case for Scalars with pointers. So far this will work for Numbers. With Strings (still there are no current examples)
                            # We will need something different to test if the value look like a pointer: regex? ####.text
                            except ValueError as e:
                                newPointer = pwobj.Pointer()
                                _setPointer(newPointer, protDict[paramName])
                                attr.setPointer(newPointer)

                self.mapper.store(prot)

        f.close()
        self.mapper.commit()

        return newDict

    def saveProtocol(self, protocol):
        self._checkModificationAllowed([protocol], 'Cannot SAVE protocol')

        if (protocol.isRunning() or protocol.isFinished()
                or protocol.isLaunched()):
            raise ModificationNotAllowedException('Cannot SAVE a protocol that is %s. '
                            'Copy it instead.' % protocol.getStatus())

        protocol.setStatus(pwprot.STATUS_SAVED)
        if protocol.hasObjId():
            self._storeProtocol(protocol)
        else:
            self._setupProtocol(protocol)

    def getProtocol(self, protId):
        protocol = self.mapper.selectById(protId)

        if not isinstance(protocol, pwprot.Protocol):
            raise Exception('>>> ERROR: Invalid protocol id: %d' % protId)

        self._setProtocolMapper(protocol)

        return protocol

    # FIXME: this function just return if a given object exists, not
    # if it is a protocol, so it is incorrect judging by the name
    # Moreover, a more consistent name (comparing to similar methods)
    # would be: hasProtocol
    def doesProtocolExists(self, protId):
        return self.mapper.exists(protId)

    def getProtocolsByClass(self, className):
        return self.mapper.selectByClass(className)

    def getObject(self, objId):
        """ Retrieve an object from the db given its id. """
        return self.mapper.selectById(objId)

    def _setHostConfig(self, protocol):
        """ Set the appropriate host config to the protocol
        give its value of 'hostname'
        """
        hostName = protocol.getHostName()
        hostConfig = self.getHostConfig(hostName)
        protocol.setHostConfig(hostConfig)

    def _storeProtocol(self, protocol):
        # Read only mode
        if not self.openedAsReadOnly():
            self.mapper.store(protocol)
            self.mapper.commit()

    def _setProtocolMapper(self, protocol):
        """ Set the project and mapper to the protocol. """

        # Tolerate loading errors. For support.
        # When only having the sqlite, sometime there are exceptions here
        # due to the absence of a set.
        from pyworkflow.mapper.sqlite import SqliteFlatMapperException
        try:

            protocol.setProject(self)
            protocol.setMapper(self.mapper)
            self._setHostConfig(protocol)

        except SqliteFlatMapperException:
            protocol.addSummaryWarning(
                "*Protocol loading problem*: A set related to this "
                "protocol couldn't be loaded.")

    def _setupProtocol(self, protocol):
        """Insert a new protocol instance in the database"""

        # Read only mode
        if not self.openedAsReadOnly():
            self._storeProtocol(protocol)  # Store first to get a proper id
            # Set important properties of the protocol
            workingDir = self.getProtWorkingDir(protocol)
            self._setProtocolMapper(protocol)

            protocol.setWorkingDir(self.getPath(PROJECT_RUNS, workingDir))
            # Update with changes
            self._storeProtocol(protocol)

    @staticmethod
    def getProtWorkingDir(protocol):
        """
        Return the protocol working directory
        """
        return "%06d_%s" % (protocol.getObjId(), protocol.getClassName())

    def getRuns(self, iterate=False, refresh=True, checkPids=False):
        """ Return the existing protocol runs in the project. 
        """
        if self.runs is None or refresh:
            # Close db open connections to db files
            if self.runs is not None:
                for r in self.runs:
                    r.closeMappers()

            # Use new selectAll Batch
            # self.runs = self.mapper.selectAll(iterate=False,
            #               objectFilter=lambda o: isinstance(o, pwprot.Protocol))
            self.runs = self.mapper.selectAllBatch(objectFilter=lambda o: isinstance(o, pwprot.Protocol))

            # Invalidate _runsGraph because the runs are updated
            self._runsGraph = None

            for r in self.runs:

                self._setProtocolMapper(r)

                # Check for run warnings
                r.checkSummaryWarnings()

                # Update nodes that are running and were not invoked
                # by other protocols
                if r.isActive():
                    if not r.isChild():
                        self._updateProtocol(r, checkPid=checkPids)

                self._annotateLastRunTime(r.endTime)

            self.mapper.commit()

        return self.runs

    def _annotateLastRunTime(self, protLastTS):
        """ Sets _lastRunTime for the project if it is after current _lastRunTime"""
        try:
            if protLastTS is None:
                return

            if self._lastRunTime is None:
                self._lastRunTime = protLastTS
            elif self._lastRunTime.datetime() < protLastTS.datetime():
                self._lastRunTime = protLastTS
        except Exception as e:
            return

    def needRefresh(self):
        """ True if any run is active and its timestamp is older than its
        corresponding runs.db
        NOTE: If an external script changes the DB this will fail. It uses
        only in memory objects."""
        for run in self.runs:
            if run.isActive():
                if not pwprot.isProtocolUpToDate(run):
                    return True
        return False

    def checkPid(self, protocol):
        """ Check if a running protocol is still alive or not.
        The check will only be done for protocols that have not been sent
        to a queue system.
        """
        from pyworkflow.protocol.launch import _runsLocally
        pid = protocol.getPid()

        if (protocol.isRunning() and _runsLocally(protocol)
            and not protocol.useQueue()
                and not pwutils.isProcessAlive(pid)):
            protocol.setFailed("Process %s not found running on the machine. "
                               "It probably has died or been killed without "
                               "reporting the status to Scipion. Logs might "
                               "have information about what happened to this "
                               "process." % pid)

    def iterSubclasses(self, classesName, objectFilter=None):
        """ Retrieve all objects from the project that are instances
            of any of the classes in classesName list.
        Params: 
            classesName: String with commas separated values of classes name. 
            objectFilter: a filter function to discard some of the retrieved
            objects."""
        for objClass in classesName.split(","):
            for obj in self.mapper.selectByClass(objClass.strip(), iterate=True,
                                                 objectFilter=objectFilter):
                yield obj

    def getRunsGraph(self, refresh=False, checkPids=False):
        """ Build a graph taking into account the dependencies between
        different runs, ie. which outputs serves as inputs of other protocols. 
        """

        if refresh or self._runsGraph is None:
            runs = [r for r in self.getRuns(refresh=refresh, checkPids=checkPids)
                    if not r.isChild()]
            self._runsGraph = self.getGraphFromRuns(runs)

        return self._runsGraph

    def getGraphFromRuns(self, runs):
        """ This function will build a dependencies graph from a set
         of given runs.
        :param runs: The input runs to build the graph
        :return: The graph taking into account run dependencies
        """
        outputDict = {}  # Store the output dict
        g = pwutils.Graph(rootName='PROJECT')

        for r in runs:
            n = g.createNode(r.strId())
            n.run = r
            n.setLabel(r.getRunName())
            outputDict[r.getObjId()] = n
            for _, attr in r.iterOutputAttributes():
                # mark this output as produced by r
                outputDict[attr.getObjId()] = n

        def _checkInputAttr(node, pointed):
            """ Check if an attr is registered as output"""
            if pointed is not None:
                pointedId = pointed.getObjId()

                if pointedId in outputDict:
                    parentNode = outputDict[pointedId]
                    if parentNode is node:
                        print("WARNING: Found a cyclic dependence from node "
                              "%s to itself, probably a bug. " % pointedId)
                    else:
                        parentNode.addChild(node)
                        return True
            return False

        for r in runs:
            node = g.getNode(r.strId())
            for _, attr in r.iterInputAttributes():
                if attr.hasValue():
                    pointed = attr.getObjValue()
                    # Only checking pointed object and its parent, if more
                    # levels we need to go up to get the correct dependencies
                    if not _checkInputAttr(node, pointed):
                        parent = self.mapper.getParent(pointed)
                        _checkInputAttr(node, parent)
        rootNode = g.getRoot()
        rootNode.run = None
        rootNode.label = "PROJECT"

        for n in g.getNodes():
            if n.isRoot() and n is not rootNode:
                rootNode.addChild(n)
        return g

    def _getRelationGraph(self, relation=pwobj.RELATION_SOURCE, refresh=False):
        """ Retrieve objects produced as outputs and
        make a graph taking into account the SOURCE relation. """
        relations = self.mapper.getRelationsByName(relation)
        g = pwutils.Graph(rootName='PROJECT')
        root = g.getRoot()
        root.pointer = None
        runs = self.getRuns(refresh=refresh)

        for r in runs:
            for paramName, attr in r.iterOutputAttributes():
                p = pwobj.Pointer(r, extended=paramName)
                node = g.createNode(p.getUniqueId(), attr.getNameId())
                node.pointer = p
                # The following alias if for backward compatibility
                p2 = pwobj.Pointer(attr)
                g.aliasNode(node, p2.getUniqueId())

        for rel in relations:
            pObj = self.getObject(rel[OBJECT_PARENT_ID])

            # Duplicated ...
            if pObj is None:
                print("WARNING: Relation seems to point to a deleted object. "
                      "%s: %s" % (OBJECT_PARENT_ID, rel[OBJECT_PARENT_ID]))
                continue

            pExt = rel['object_parent_extended']
            pp = pwobj.Pointer(pObj, extended=pExt)

            if pObj is None or pp.get() is None:
                print("project._getRelationGraph: ERROR, pointer to parent is "
                      "None. IGNORING IT.\n")
                for key in rel.keys():
                    print("%s: %s" % (key, rel[key]))

                continue

            pid = pp.getUniqueId()
            parent = g.getNode(pid)

            while not parent and pp.hasExtended():
                pp.removeExtended()
                parent = g.getNode(pp.getUniqueId())

            if not parent:
                print("project._getRelationGraph: ERROR, parent Node "
                      "is None: ", pid)
            else:
                cObj = self.getObject(rel['object_child_id'])
                cExt = rel['object_child_extended']

                if cObj is not None:
                    if cObj.isPointer():
                        cp = cObj
                        if cExt:
                            cp.setExtended(cExt)
                    else:
                        cp = pwobj.Pointer(cObj, extended=cExt)
                    child = g.getNode(cp.getUniqueId())

                    if not child:
                        print("project._getRelationGraph: ERROR, child Node "
                              "is None: ", cp.getUniqueId())
                        print("   parent: ", pid)
                    else:
                        parent.addChild(child)
                else:
                    print("project._getRelationGraph: ERROR, child Obj "
                          "is None, id: ", rel['object_child_id'])
                    print("   parent: ", pid)

        for n in g.getNodes():
            if n.isRoot() and n is not root:
                root.addChild(n)

        return g

    def getSourceChilds(self, obj):
        """ Return all the objects have used obj
        as a source.
        """
        return self.mapper.getRelationChilds(pwobj.RELATION_SOURCE, obj)

    def getSourceParents(self, obj):
        """ Return all the objects that are SOURCE of this object.
        """
        return self.mapper.getRelationParents(pwobj.RELATION_SOURCE, obj)

    def getTransformGraph(self, refresh=False):
        """ Get the graph from the TRANSFORM relation. """
        if refresh or not self._transformGraph:
            self._transformGraph = self._getRelationGraph(pwobj.RELATION_TRANSFORM,
                                                          refresh)

        return self._transformGraph

    def getSourceGraph(self, refresh=False):
        """ Get the graph from the SOURCE relation. """
        if refresh or not self._sourceGraph:
            self._sourceGraph = self._getRelationGraph(pwobj.RELATION_SOURCE,
                                                       refresh)

        return self._sourceGraph

    def getRelatedObjects(self, relation, obj, direction=pwobj.RELATION_CHILDS,
                          refresh=False):
        """ Get all objects related to obj by a give relation.
        Params:
            relation: the relation name to search for.
            obj: object from which the relation will be search,
                actually not only this, but all other objects connected
                to this one by the pwobj.RELATION_TRANSFORM.
            direction: this say if search for childs or parents in the relation.
        """
        graph = self.getTransformGraph(refresh)
        relations = self.mapper.getRelationsByName(relation)
        connection = self._getConnectedObjects(obj, graph)

        objects = []
        objectsDict = {}

        for rel in relations:
            pObj = self.getObject(rel[OBJECT_PARENT_ID])

            if pObj is None:
                print("WARNING: Relation seems to point to a deleted object. "
                      "%s: %s" % (OBJECT_PARENT_ID, rel[OBJECT_PARENT_ID]))
                continue
            pExt = rel['object_parent_extended']
            pp = pwobj.Pointer(pObj, extended=pExt)

            if pp.getUniqueId() in connection:
                cObj = self.getObject(rel['object_child_id'])
                cExt = rel['object_child_extended']
                cp = pwobj.Pointer(cObj, extended=cExt)
                if cp.hasValue() and cp.getUniqueId() not in objectsDict:
                    objects.append(cp)
                    objectsDict[cp.getUniqueId()] = True

        return objects

    def _getConnectedObjects(self, obj, graph):
        """ Given a TRANSFORM graph, return the elements that
        are connected to an object, either childs, ancestors or siblings. 
        """
        n = graph.getNode(obj.strId())
        # Get the oldest ancestor of a node, before reaching the root node
        while n is not None and not n.getParent().isRoot():
            n = n.getParent()

        connection = {}

        if n is not None:
            # Iterate recursively all descendants
            for node in n.iterChilds():
                connection[node.pointer.getUniqueId()] = True
                # Add also 
                connection[node.pointer.get().strId()] = True

        return connection

    def isReadOnly(self):
        if getattr(self, 'settings', None) is None:
            return False

        return self.settings.getReadOnly()

    def isInReadOnlyFolder(self):
        return self._isInReadOnlyFolder

    def openedAsReadOnly(self):
        return self.isReadOnly() or self.isInReadOnlyFolder()

    def setReadOnly(self, value):
        self.settings.setReadOnly(value)

    def fixLinks(self, searchDir):
        print("Fixing project links. Searching at %s" % searchDir)
        runs = self.getRuns()

        for prot in runs:
            print (prot)
            broken = False
            if isinstance(prot, ProtImportBase) or prot.getClassName() == "ProtImportMovies":
                print("Import detected")
                for _, attr in prot.iterOutputAttributes():
                    fn = attr.getFiles()
                    for f in attr.getFiles():
                        if ':' in f:
                            f = f.split(':')[0]

                        if not os.path.exists(f):
                            if not broken:
                                broken = True
                                print("Found broken links in run: ",
                                      pwutils.magenta(prot.getRunName()))
                            print("  Missing: ", pwutils.magenta(f))
                            if os.path.islink(f):
                                print("    -> ", pwutils.red(os.path.realpath(f)))
                            newFile = pwutils.findFile(os.path.basename(f),
                                                       searchDir,
                                                       recursive=True)
                            if newFile:
                                print("  Found file %s, creating link..." % newFile,
                                      pwutils.green("   %s -> %s" % (f, newFile)))
                                pwutils.createAbsLink(newFile, f)

    @staticmethod
    def cleanProjectName(projectName):
        """ Cleans a project name to avoid common errors
        Use it whenever you want to get the final project name pyworkflow will endup.
        Spaces will be replaced by _ """
        return projectName.replace(" ", "_")


class MissingProjectDbException(Exception):
    pass


class ModificationNotAllowedException(Exception):
    pass



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Pablo Conesa [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This modules hosts code provider and window to search for a run"""

import tkinter as tk
from pyworkflow import Config
import pyworkflow.protocol as pwprot

from pyworkflow.gui import SearchBaseWindow
from pyworkflow.gui.project.constants import ACTION_COPY, ACTION_EDIT, ACTION_RENAME, ACTION_DELETE, ACTION_STEPS, \
    ACTION_BROWSE, ACTION_DB, ACTION_STOP, ACTION_EXPORT, ACTION_EXPORT_UPLOAD, ACTION_COLLAPSE, ACTION_EXPAND, \
    ACTION_LABELS, ACTION_SELECT_FROM, ACTION_SELECT_TO, ACTION_RESTART_WORKFLOW, ACTION_CONTINUE_WORKFLOW, \
    ACTION_STOP_WORKFLOW, ACTION_RESET_WORKFLOW, ActionIcons

from pyworkflow.gui.tree import ProjectRunsTreeProvider

class RunsTreeProvider(ProjectRunsTreeProvider):
    """Provide runs info to populate tree inside viewprotocols. Is more advanced
    than ProjectRunsTreeProvider extended, extended with right click actions."""

    def __init__(self, project, actionFunc):
        super().__init__(project)
        self.actionFunc = actionFunc
        self._selection = project.getSettings().runSelection

    def getActionsFromSelection(self):
        """ Return the list of options available for selection. """
        n = len(self._selection)
        single = n == 1
        if n:
            prot = self.project.getProtocol(self._selection[0])
            status = prot.getStatus()
            nodeInfo = self.project.getSettings().getNodeById(prot.getObjId())
            expanded = nodeInfo.isExpanded() if nodeInfo else True
        else:
            status = None

        stoppable = status in [pwprot.STATUS_RUNNING, pwprot.STATUS_SCHEDULED,
                               pwprot.STATUS_LAUNCHED]

        return [(ACTION_EDIT, single and status and expanded),
                (ACTION_RENAME, single and status and expanded),
                (ACTION_COPY, status and expanded),
                (ACTION_DELETE, status != pwprot.STATUS_RUNNING and status and expanded),
                (ACTION_STEPS, single and Config.debugOn() and status and expanded),
                (ACTION_BROWSE, single and status and expanded),
                (ACTION_DB, single and Config.debugOn() and status and expanded),
                (ACTION_STOP, stoppable and single),
                (ACTION_EXPORT, not single),
                (ACTION_EXPORT_UPLOAD, not single),
                (ACTION_COLLAPSE, single and status and expanded),
                (ACTION_EXPAND, single and status and not expanded),
                (ACTION_LABELS, True),
                (ACTION_SELECT_FROM, True),
                (ACTION_SELECT_TO, True),
                (ACTION_RESTART_WORKFLOW, single),
                (ACTION_CONTINUE_WORKFLOW, single),
                (ACTION_STOP_WORKFLOW, single),
                (ACTION_RESET_WORKFLOW, single)
                ]

    def getObjectActions(self, obj):

        def addAction(actionLabel):
            if actionLabel:
                text = actionLabel
                action = actionLabel
                actionLabel = (text, lambda: self.actionFunc(action),
                               ActionIcons.get(action, None))
            return actionLabel

        actions = [addAction(a)
                   for a, cond in self.getActionsFromSelection() if cond]

        if hasattr(obj, 'getActions'):
            for text, action in obj.getActions():
                actions.append((text, action, None))

        return actions


class SearchRunWindow(SearchBaseWindow):

    columnConfig = {
        '#0': (ProjectRunsTreeProvider.ID_COLUMN, {'width': 100, 'stretch': tk.NO}, 10),
        ProjectRunsTreeProvider.RUN_COLUMN: (ProjectRunsTreeProvider.RUN_COLUMN, {'width': 300, 'stretch': tk.TRUE}, 10),
        ProjectRunsTreeProvider.STATE_COLUMN: (ProjectRunsTreeProvider.STATE_COLUMN, {'width': 150, 'stretch': tk.FALSE}, 5),
        ProjectRunsTreeProvider.TIME_COLUMN: (ProjectRunsTreeProvider.TIME_COLUMN, {'width': 200, 'stretch': tk.FALSE}, 5),
        'Comment': ('Comment', {'width': 300, 'stretch': tk.FALSE}, 5),
        'Expanded': ('Expanded', {'width': 150, 'stretch': tk.FALSE}, 5),
    }

    def __init__(self, parentWindow, runsGraph, **kwargs):


        super().__init__(parentWindow,
                         title="Locate a protocol in the graph",
                         **kwargs)
        self.runsGraph = runsGraph

    def _onSearchClick(self, e=None):

        self._resultsTree.clear()
        keyword = self._searchVar.get().lower().strip()

        weightIndex = len(self.columnConfig)
        nodes = self.runsGraph.getNodes()
        protList = []

        for node in nodes:
            if node.run is not None:
                run = node.run
                key = run.getObjId()
                label = run.getRunName()
                status = run.getStatusMessage(),
                time = run.getObjCreation()
                comment = run.getObjComment()
                expanded = "expanded" if getattr(node, 'expanded', False) else "collapsed"
                line = (key, label, status, time, comment, expanded)

                line = self.addSearchWeight(line, keyword)
                # something was found: weight > 0
                if line[weightIndex] != 0:
                    # Add the run
                    protList.append(line)

        # Sort by weight
        protList.sort(reverse=True, key=lambda x: x[weightIndex])

        for line in protList:

            self._resultsTree.insert(
                '', 'end', line[0], text=line[0],
                values=line[1:])






# **************************************************************************
# *
# * Authors:      Roberto Marabini (roberto@cnb.csic.es) [1]
# *               J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import re


class ProgressBar(object):
    """ Text progress bar class for Python.

    A text progress bar is typically used to display the progress of a long
    running operation, providing a visual cue that processing is underway.

    Example:
        N = 1000
        pb = ProgressBar(N, fmt=ProgressBar.FULL)
        pb.start()
        for x in range(N):
            pb.update(x+1)
            sleep(0.1)
        pb.finish()
    """
    DEFAULT = 'Progress: %(bar)s %(percent)3d%%'
    FULL = '%(bar)s %(current)d/%(total)d (%(percent)3d%%) %(remaining)d to go'
    # scipion uses variable fonts so the bar size changes
    # since the space width is different from the = width
    NOBAR = '%(current)d/%(total)d (%(percent)3d%%) %(remaining)d to go'
    OBJID = '%(bar)s %(current)d/%(total)d (%(percent)3d%%) (objectId=%(objectId)d)'
    DOT = '.'

    def __init__(self, total, width=40, fmt=DEFAULT, symbol='=',
                 output=None, extraArgs=None):
        """
        Create a new ProgressBar object.

        :param total: The total amount that will be running the progress bar.
            The value in the update() method can no go greater than this value.
        :param width: progress bar width (without the percentage and number of
            iterations loop)
        :param fmt: predefined format string, so far DEFAULT, FULL, OBJID and
            DOT are defined.
        :param symbol: progress bar is made with this symbol
        :param output:
        :param extraArgs: Additional arguments that can be passed to be used
            the fmt format. (e.g extraArgs={'objectId': 1} for fmt=OBJID
        """
        if len(symbol) != 1:
            raise Exception("Symbol should be only 1 character length. ")

        self._total = total
        self._width = width
        self._symbol = symbol
        self._output = output or sys.stdout
        self._current = -1
        self._extraArgs = extraArgs or {}
        self._fmt = fmt
        self._directPrint = fmt == self.DOT

        if not self._directPrint:
            # This line computes the number of digits
            # in total and rewrites the fmt string
            # so if total = 100, %d is converted to %3d
            self._fmt = re.sub(r'(?P<name>%\(.+?\))d',
                               r'\g<name>%dd' % len(str(total)), fmt)

    def __getStr(self):
        """ Internal function to return the current string value.
        It should be called after the value has being set.
        """
        if self._directPrint:  # print just a dot
            return self._fmt if self._current else ''

        percent = self._current / float(self._total)
        size = int(self._width * percent)
        remaining = self._total - self._current
        bar = '[' + self._symbol * size + ' ' * (self._width - size) + ']'

        args = {
            'total': self._total,
            'bar': bar,
            'current': self._current,
            'percent': percent * 100,
            'remaining': remaining,
        }
        args.update(self._extraArgs)

        return '\r' + self._fmt % args

    def start(self):
        """ Print empty progress bar. """
        self.update(0)

    def update(self, value):
        """
        Update the current value and print the progress.

        :param value: New value, should be greater than the previous
            value and less or equal the total value
        :return:
        """
        if value < 0 or value <= self._current or value > self._total:
            raise Exception("Incorrect value provided. It should be greater "
                            "than previous value and between 0 and total. ")
        self._current = value
        self._output.write(self.__getStr())
        self._output.flush()

    def finish(self, printNewLine=True):
        """ Finalize the progress and
        print last update with 100% complete message """
        if self._current < self._total:
            self.update(self._total)
        # print a new line
        if printNewLine:
            self._output.write("\n")



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Text based widgets.
"""


import os
import sys
import time
import webbrowser
import subprocess
import tkinter.ttk as ttk
import tkinter as tk
import tkinter.messagebox as tkMessageBox

import pyworkflow as pw
from pyworkflow.utils import (HYPER_BOLD, HYPER_ITALIC, HYPER_LINK1, HYPER_LINK2,
                              parseHyperText, renderLine, renderTextFile, colorName,
                              which, envVarOn, expandPattern)
from pyworkflow.utils.properties import Message, Color, Icon
from . import gui
from .widgets import Scrollable, IconButton
from .tooltip import ToolTip


# Define a function to open files cleanly in a system-dependent way
if sys.platform.startswith('darwin'):  # macs use the "open" command
    def _open_cmd(path, tkParent=None):
        subprocess.Popen(['open', path])
elif os.name == 'nt':  # there is a function os.startfile for windows
    def _open_cmd(path, tkParent=None):
        os.startfile(path)
elif os.name == 'posix':  # linux systems and so on
    def find_prog(*args):
        """Return the first argument that is a program in PATH"""
        for command in args:
            if which(command):
                return command
        return None

    x_open = find_prog('xdg-open', 'gnome-open', 'kde-open', 'gvfs-open')
    editor = find_prog('pluma', 'gedit', 'kwrite', 'geany', 'kate',
                       'emacs', 'nedit', 'mousepad', 'code')

    def _open_cmd(path, tkParent=None):
        # If it is an url, open with browser.
        if path.startswith('http://') or path.startswith('https://') or path.endswith('.html'):
            try:
                webbrowser.open_new_tab(path)
                return
            except:
                pass
        # OK, it is a file. Check if it does exist
        # and notify if it does not
        if not os.path.isfile(path):
            try:
                # if tkRoot is null the error message may be behind
                # other windows
                tkMessageBox.showerror("File Error",  # bar title
                                       "File not found\n(%s)" % path,  # message
                                       parent=tkParent)
                return
            except:
                return

        if x_open:  # standard way to open
            proc = subprocess.Popen([x_open, path])
            time.sleep(1)
            if proc.poll() in [None, 0]:
                return  # yay! that's the way to do it!
        if editor:  # last card: try to open it in an editor
            proc = subprocess.Popen([editor, path])
            time.sleep(1)
            if proc.poll() in [None, 0]:
                return  # hope we found your fav editor :)
        print('WARNING: Cannot open %s' % path)  # nothing worked! :(
else:
    def _open_cmd(path, tkParent=None):
        try:
            tkMessageBox.showerror("Unknown System",  # bar title
                                   'Unknown system, so cannot open %s' % path,  # message
                                   parent=tkParent)
            return
        except:
            pass


class HyperlinkManager:
    """ Tkinter Text Widget Hyperlink Manager, taken from:
    http://effbot.org/zone/tkinter-text-hyperlink.htm """
    def __init__(self, text):
        self.text = text
        self.text.tag_config("hyper", foreground=Color.RED_COLOR, underline=1)
        self.text.tag_bind("hyper", "<Enter>", self._enter)
        self.text.tag_bind("hyper", "<Leave>", self._leave)
        self.text.tag_bind("hyper", "<Button-1>", self._click)
        self.reset()

    def reset(self):
        self.links = {}

    def add(self, action):
        # add an action to the manager.  returns tags to use in
        # associated text widget
        tag = "hyper-%d" % len(self.links)
        self.links[tag] = action
        return "hyper", tag

    def _enter(self, event):
        self.text.config(cursor="hand2")

    def _leave(self, event):
        self.text.config(cursor="")

    def _click(self, event):
        for tag in self.text.tag_names(tk.CURRENT):
            if tag[:6] == "hyper-":
                self.links[tag]()
                return


class Text(tk.Text, Scrollable):    
    """ Base Text widget with some functionality
    that will be used by children classes.
    """
    def __init__(self, master, **opts):
        if 'handlers' in opts:
            self.handlers = opts.pop('handlers')
        else:
            self.handlers = {}
        opts['font'] = gui.fontNormal
        defaults = self.getDefaults()
        defaults.update(opts)
        Scrollable.__init__(self, master, tk.Text, wrap=tk.WORD, **opts)
        self._createWidgets(master, **defaults)
        self.configureTags()

    def _createWidgets(self, master, **opts):
        """This is an internal function to create the Text, the Scrollbar and the Frame"""
        
        # create a popup menu
        self.menu = tk.Menu(master, tearoff=0, postcommand=self.updateMenu)
        self.menu.add_command(label="Copy to clipboard", command=self.copyToClipboard)
        self.menu.add_command(label="Open path", command=self.openFile)
        # Associate with right click
        self.bind("<Button-1>", self.onClick)
        self.bind("<Button-3>", self.onRightClick)
        
    def getDefaults(self):
        """This should be implemented in subclasses to provide defaults"""
        return {}
    
    def configureTags(self):
        """This should be implemented to create specific tags"""
        pass
    
    def addLine(self, line):
        """Should be implemented to add a line """
        self.insert(tk.END, line + '\n')
        
    def addNewline(self):
        self.insert(tk.END, '\n')
        
    def goBegin(self):
        self.see(0.0)
        
    def goEnd(self):
        self.see(tk.END)
        
    def isAtEnd(self):
        return self.scrollbar.get() == 1.0
        
    def clear(self):
        self.delete(0.0, tk.END)

    def getText(self):

        textWithNewLine = self.get(0.0, tk.END)

        # Remove the last new line
        return textWithNewLine.rstrip('\n')

    def setText(self, text):
        """ Replace the current text with new one. """
        self.clear()
        self.addText(text)
        
    def addText(self, text):
        """ Add some text to the current state. """
        if isinstance(text, list):
            for line in text:
                self.addLine(line)
        else:
            for line in text.splitlines():
                self.addLine(line)
        
    def onClick(self, e=None): 
        self.selection = None
        self.selection_clear()
        self.menu.unpost()
        
    def onRightClick(self, e):
        try:
            self.selection = self.selection_get().strip()
            self.menu.post(e.x_root, e.y_root)    
        except tk.TclError as e:
            pass
    
    def copyToClipboard(self, e=None):
        self.clipboard_clear()
        self.clipboard_append(self.selection)

    def openFile(self):
        # What happens when you right-click and select "Open path"
        self.openPath(self.selection)

    def openPath(self, path):
        """Try to open the selected path"""
        path = expandPattern(path)

        # If the path is a dir, open it with scipion browser dir <path>
        if os.path.isdir(path):
            dpath = (path if os.path.isabs(path)
                     else os.path.join(os.getcwd(), path))
            subprocess.Popen([pw.PYTHON, pw.getViewerScript(), dpath])
            return

        # If it is a file, interpret it correctly and open it with DataView
        dirname = os.path.dirname(path)
        fname = os.path.basename(path)
        if '@' in fname:
            path = os.path.join(dirname, fname.split('@', 1)[-1])
        else:
            path = os.path.join(dirname, fname)

        if os.path.exists(path):
            from pwem import emlib
            fn = emlib.FileName(path)
            if fn is not None and (fn.isImage() or fn.isMetaData()):
                # fn is None if xmippLib is the xmippLib ghost library
                from pwem.viewers import DataView
                DataView(path).show()
            else:
                _open_cmd(path)
        else:
            # This is probably one special reference, like sci-open:... that
            # can be interpreted with our handlers.
            tag = path.split(':', 1)[0] if ':' in path else None
            if tag in self.handlers:
                self.handlers[tag](path.split(':', 1)[-1])
            else:
                print("Can't find %s" % path)

    def updateMenu(self, e=None):
        state = 'normal'
        # if not xmippExists(self.selection):
        #    state = 'disabled'#self.menu.entryconfig(1, background="green")
        self.menu.entryconfig(1, state=state)
        
    def setReadOnly(self, value):
        state = tk.NORMAL
        if value:
            state = tk.DISABLED
        self.config(state=state) 
        
    def highlight(self, pattern, tag, start="1.0", end="end", regexp=False):
        """ Apply the given tag to all text that matches the given pattern

        If 'regexp' is set to True, pattern will be treated as a regular expression
        Taken from: 
            http://stackoverflow.com/questions/3781670/tkinter-text-highlighting-in-python
        """
        start = self.index(start)
        end = self.index(end)
        self.mark_set("matchStart", start)
        self.mark_set("matchEnd", start)
        self.mark_set("searchLimit", end)

        count = tk.IntVar()
        while True:
            index = self.search(pattern, "matchEnd", "searchLimit",
                                count=count, regexp=regexp)
            if index == "":
                break
            self.mark_set("matchStart", index)
            self.mark_set("matchEnd", "%s+%sc" % (index, count.get()))
            self.tag_add(tag, "matchStart", "matchEnd")


def configureColorTags(text):
    """ Create tags in text (of type tk.Text) for all the supported colors. """
    try:
        for color in colorName.values():
            text.tag_config(color, foreground=color)
        return True
    except Exception as e:
        print("Colors still not available (%s)" % e)
        return False

       
class TaggedText(Text):  
    """
    Implement a Text that will recognize some basic tags
    *some_text* will display some_text in bold
    _some_text_ will display some_text in italic
    some_link or [[some_link][some_label]] will display some_link
     as hyperlink or some_label as hyperlink to some_link
    also colors are recognized if set option colors=True
    """           
    def __init__(self, master, colors=True, **opts):  
        self.colors = colors
        Text.__init__(self, master, **opts)
        self.hm = HyperlinkManager(self)

    def getDefaults(self):
        return {'bg': "white", 'bd': 0}
        # It used to have also 'font': gui.fontNormal  but that stops
        # this file from running. Apparently there is no fontNormal in gui.

    def configureTags(self):
        self.tag_config('normal', justify=tk.LEFT, font=gui.fontNormal)
        self.tag_config(HYPER_BOLD, justify=tk.LEFT, font=gui.fontBold)
        self.tag_config(HYPER_ITALIC, justify=tk.LEFT, font=gui.fontItalic)
        if self.colors:            
            self.colors = configureColorTags(self)
            # Color can be unavailable, so disable use of colors
        
    @staticmethod
    def openLink(link):
        webbrowser.open_new_tab(link)  # Open in the same browser, new tab

    @staticmethod
    def mailTo(email):
        webbrowser.open("mailto:" + email)

    def matchHyperText(self, match, tag):
        """ Process when a match a found and store indexes inside string."""
        self.insert(tk.END, self.line[self.lastIndex:match.start()])
        g1 = match.group(tag)

        if tag == HYPER_BOLD or tag == HYPER_ITALIC:
            self.insert(tk.END, ' ' + g1, tag)
        elif tag == HYPER_LINK1:
            self.insert(tk.END, g1, self.hm.add(lambda: self.openLink(g1)))
        elif tag == HYPER_LINK2:
            label = match.group('link2_label')
            if g1.startswith('http'):
                self.insert(tk.END, label, self.hm.add(lambda: self.openLink(g1)))
            elif g1.startswith('mailto:'):
                self.insert(tk.END, label, self.hm.add(lambda: self.mailTo(g1)))
            else:
                self.insert(tk.END, label, self.hm.add(lambda: self.openPath(g1)))
        self.lastIndex = match.end()
        
        return g1

    def addLine(self, line):
        self.line = line
        self.lastIndex = 0
        if line is not None:
            parseHyperText(line, self.matchHyperText)
            Text.addLine(self, line[self.lastIndex:])


class OutputText(Text):
    """
    Implement a Text that will show file content
    and handle console metacharacter for colored output
    """
    def __init__(self, master, filename, colors=True, t_refresh=0, maxSize=400, **opts):
        """ colors flag indicate if try to parse color meta-characters
            t_refresh is the refresh time in seconds, 0 means no refresh
        """
        self.filename = filename
        self.colors = colors
        self.t_refresh = t_refresh
        self.maxSize = maxSize

        self.refreshAlarm = None  # Identifier returned by after()
        self.lineNo = 0
        self.offset = 0
        self.lastLine = ''
        Text.__init__(self, master, **opts)
        self.hm = HyperlinkManager(self)
        self.doRefresh()

    def getDefaults(self):
        return {'bg': "black", 'fg': 'white', 'bd': 0,
                'height': 30,  'width': 100}
        # It used to have also  'font': gui.fontNormal  but that stops this
        # file from running. Apparently there is no fontNormal in gui.
        
    def configureTags(self):
        if self.colors:
            configureColorTags(self)

    def _removeLastLine(self):
        line = int(self.index(tk.END).split('.')[0])
        if line > 0:
            line -= 1
            self.delete('%d.0' % line, tk.END)
        
    def addLine(self, line):
        renderLine(line, self._addChunk, self.lineNo)

    def _addChunk(self, txt, fmt=None):
        """
        Add text txt to the widget, with format fmt.
        fmt can be a color (like 'red') or a link that looks like 'link:url'.
        """
        if self.colors and fmt is not None:
            if fmt.startswith('link:'):
                fname = fmt.split(':', 1)[-1]
                self.insert(tk.END, txt, self.hm.add(lambda: openTextFileEditor(fname)))
            else:
                self.insert(tk.END, txt, fmt)
        else:
            self.insert(tk.END, txt)

    def _notifyLine(self, line):
        if '\r' in self.lastLine and '\r' in line:
            self._removeLastLine()
            self.addNewline()
            
        self.lastLine = line        
        
    def readFile(self, clear=False):
        self.setReadOnly(False)
        
        if clear:
            self.offset = 0
            self.lineNo = 0
            self.clear()

        if os.path.exists(self.filename):
            self.offset, self.lineNo = renderTextFile(self.filename, 
                                                      self._addChunk,
                                                      offset=self.offset, 
                                                      lineNo=self.lineNo,
                                                      maxSize=self.maxSize, 
                                                      notifyLine=self._notifyLine,
                                                      errors='replace')
        else:
            self.insert(tk.END, "File '%s' doesn't exist" % self.filename)

        self.setReadOnly(True)
        # self.goEnd()
      
    def doRefresh(self):
        # First stop pending refreshes
        if self.refreshAlarm:
            self.after_cancel(self.refreshAlarm)
            self.refreshAlarm = None

        self.readFile()

        if self.t_refresh > 0:
            self.refreshAlarm = self.after(self.t_refresh*1000, self.doRefresh)


class TextFileViewer(tk.Frame):
    """ Implementation of a simple text file viewer """
    
    LabelBgColor = "white"
    
    def __init__(self, master, fileList=[],
                 allowSearch=True, allowRefresh=True, allowOpen=False,
                 font=None, maxSize=400, width=100, height=30):
        tk.Frame.__init__(self, master)
        self.searchList = None
        self.lastSearch = None
        self.refreshAlarm = None
        self._lastTabIndex = None
        self.fileList = []  # Files being visualized
        self.taList = []  # Text areas (OutputText, a scrollable TkText)
        self.fontDict = {}
        self._allowSearch = allowSearch
        self._allowRefresh = allowRefresh
        self._allowOpen = allowOpen
        self._font = font  # allow a font to be passed as argument to be used
        self.maxSize = maxSize
        self.width = width
        self.height = height

        self.createWidgets(fileList)
        self.master = master
        self.addBinding()
        
    def addFile(self, filename):
        self.fileList.append(filename)
        self._addFileTab(filename)
        
    def clear(self):
        """ Remove all added files. """
        self.fileList = []
        for _ in self.taList:
            self.notebook.forget(0)       
        self.taList = []
        self._lastTabIndex = None
        
    def _addFileTab(self, filename):
        tab = tk.Frame(self.notebook)
        tab.rowconfigure(0, weight=1)
        tab.columnconfigure(0, weight=1)
        kwargs = {'bg': 'black',
                  'fg': 'white'}
        
        if self._font is not None:
            kwargs['font'] = self._font

        t = OutputText(tab, filename, width=self.width, height=self.height,
                       maxSize=self.maxSize, **kwargs)
        t.frame.grid(column=0, row=0, padx=5, pady=5, sticky='nsew')
        self.taList.append(t)
        tabText = "   %s   " % os.path.basename(filename)
        self.notebook.add(tab, text=tabText)        
    
    def createWidgets(self, fileList):
        # registerCommonFonts()
        self.columnconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)        

        # Create toolbar frame
        toolbarFrame = tk.Frame(self)
        toolbarFrame.grid(column=0, row=0, padx=5, sticky='new')
        gui.configureWeigths(toolbarFrame)
        # Add the search box
        right = tk.Frame(toolbarFrame)
        right.grid(column=1, row=0, sticky='ne')        
        self.searchVar = tk.StringVar()
        if self._allowSearch:
            tk.Label(right, text='Search:').grid(row=0, column=3, padx=5)
            self.searchEntry = tk.Entry(right, textvariable=self.searchVar,
                                        font=self._font)
            self.searchEntry.grid(row=0, column=4, sticky='ew', padx=5)
            self.searchEntry.bind('<Return>', self.findText)
            self.searchEntry.bind('<KP_Enter>', self.findText)
            # btn = IconButton(right, "Search", Icon.ACTION_SEARCH,
            #                  tooltip=Message.TOOLTIP_SEARCH,
            #                  command=self.findText, bg=None)
            # btn.grid(row=0, column=5, padx=(0, 5))

            btn = IconButton(right, "Next", Icon.ACTION_FIND_NEXT,
                             tooltip=Message.TOOLTIP_SEARCH_NEXT,
                             command=self.findText, bg=None)
            btn.grid(row=0, column=5, padx=(0, 5))

            btn = IconButton(right, "Previous", Icon.ACTION_FIND_PREVIOUS,
                             tooltip=Message.TOOLTIP_SEARCH_PREVIOUS,
                             command=self.findPrevText, bg=None)
            btn.grid(row=0, column=6, padx=(0, 5))

        if self._allowRefresh:
            btn = IconButton(right, "Refresh", Icon.ACTION_REFRESH,
                             tooltip=Message.TOOLTIP_REFRESH,
                             command=self._onRefresh, bg=None)
            btn.grid(row=0, column=7, padx=(0, 5), pady=2)
        if self._allowOpen:
            btn = IconButton(right, "Open external", Icon.ACTION_REFERENCES,
                             tooltip=Message.TOOLTIP_EXTERNAL,
                             command=self._openExternal, bg=None)
            btn.grid(row=0, column=8, padx=(0, 5), pady=2)

        # Create tabs frame
        tabsFrame = tk.Frame(self)
        tabsFrame.grid(column=0, row=1, padx=5, pady=(0, 5), sticky="nsew")
        tabsFrame.columnconfigure(0, weight=1)
        tabsFrame.rowconfigure(0, weight=1)
        self.notebook = ttk.Notebook(tabsFrame)  
        self.notebook.rowconfigure(0, weight=1)
        self.notebook.columnconfigure(0, weight=1)      
        for f in fileList:
            self._addFileTab(f)
        self.notebook.grid(column=0, row=0, sticky='nsew', padx=5, pady=5)   
        self.notebook.bind('<<NotebookTabChanged>>', self._tabChanged)

    def _tabChanged(self, e=None):
        self._lastTabIndex = self.notebook.select()
        # reset the search
        self.lastSearch = None
        # Setting the focus, captures it when selecting protocols and
        # therefore "deleting" using keys or other future shortcut for the canvas
        # will not work.
        # self.searchEntry.focus_set()

    def addBinding(self):

        shortcutDefinitions = [(lambda e: self.findText(), "Trigger the search", ['<Return>']),
                               (lambda e: self.findText(), "Trigger the search", ['<KP_Enter>']),
                               (lambda e: self.findText(matchCase=True), "Trigger a case sensitive search", ['<Shift-Return>']),
                               (lambda e: self.findText(), "Move to the next highlighted item", ["<Down>", '<F3>']),
                               (lambda e: self.findText(-1), "Move to the previous highlighted item", ["<Up>", '<Shift-F3>']),
                               ]
        tooltip = "Shortcuts:"

        for callback, help, keys in shortcutDefinitions:
            tooltip += "\n" + help + ": "
            for key in keys:
                self.searchEntry.bind(key, callback)
                tooltip += key

        # Add a tooltip
        ToolTip(self.searchEntry, tooltip, 800)

    def getIndex(self):
        """ Return the index of the selected tab. """
        selected = self.notebook.select()
        if selected:
            return self.notebook.index(selected)
        return -1
    
    def setIndex(self, index):
        """ Select the tab with the given index. """
        if index != -1:
            self.notebook.select(self.notebook.tabs()[index])
    
    def selectedText(self):
        index = self.getIndex()
        if index != -1:
            return self.taList[index]
        return None
    
    def changeFont(self, event=""):
        for font in self.fontDict.values():
            gui.changeFontSize(font, event)
              
    def refreshAll(self, clear=False, goEnd=False):
        """ Refresh all output textareas. """
        for ta in self.taList:
            ta.readFile(clear)
            if goEnd:
                ta.goEnd()
        if self._lastTabIndex is not None:
            self.notebook.select(self._lastTabIndex)
            
    def _onRefresh(self, e=None):
        """ Action triggered when the 'Refresh' icon is clicked. """
        self.refreshAll(clear=False, goEnd=True)
        
    def refreshOutput(self, e=None):
        if self.refreshAlarm:
            self.after_cancel(self.refreshAlarm)
            self.refreshAlarm = None
        text = self.selectedText()
        if text:
            text.readFile()
        
    def changePosition(self, index):
        self.selectedText().see(index)

    def findPrevText(self):
        self.findText(-1)

    def findText(self, direction=1, matchCase=0):
        text = self.selectedText()
        str = self.searchVar.get()
        if text:
            if str is None or str != self.lastSearch:
                self.buildSearchList(text, str, matchCase=matchCase)
                self.lastSearch = str

            else:
                self.nextSearchIndex(text, direction)
            self.searchEntry.focus_set()
        
    def buildSearchList(self, text, str, matchCase=0):
        text.tag_remove('found', '1.0', tk.END)
        list = []
        if str:
            idx = '1.0'
            while True:
                idx = text.search(str, idx, nocase=not matchCase, stopindex=tk.END)
                if not idx:
                    break
                lastidx = '%s+%dc' % (idx, len(str))
                text.tag_add('found', idx, lastidx)
                list.append((idx, lastidx))
                idx = lastidx
        text.tag_config('found', foreground='white', background='blue')
        # Set class variables
        self.searchList = list
        self.currentIndex = -1
        self.nextSearchIndex(text)  # select first element
    
    def nextSearchIndex(self, text, direction=1):
        # use direction=-1 to go backward
        text.tag_remove('found_current', '1.0', tk.END)
        if len(self.searchList) == 0:
            return
        self.currentIndex = (self.currentIndex + direction) % len(self.searchList)
        idx, lastidx = self.searchList[self.currentIndex]
        text.tag_config('found_current', foreground='yellow', background='red')
        text.tag_add('found_current', idx, lastidx)
        text.see(idx)
        
    def _openExternal(self):
        """ Open a new window with an external viewer. """
        if envVarOn('SCIPION_EXTERNAL_VIEWER'):
            if not self.taList:
                return
            openTextFileEditor(self.taList[max(self.getIndex(), 0)].filename)
        else:
            showTextFileViewer("File viewer", self.fileList, self.windows)
  
  
def openTextFile(filename):
    """ Open a text file with an external or default viewer. """
    if envVarOn('SCIPION_EXTERNAL_VIEWER'):
        openTextFileEditor(filename)
    else:
        showTextFileViewer("File viewer", [filename])    
    
    
def openTextFileEditor(filename, tkParent=None):
    try:
        _open_cmd(filename, tkParent)
    except:
        showTextFileViewer("File viewer", [filename])
    
    
def showTextFileViewer(title, filelist, parent=None, main=False):
    w = gui.Window(title, parent, minsize=(600, 400))
    viewer = TextFileViewer(w.root, filelist, maxSize=-1, font=w.font)
    viewer.grid(row=0, column=0, sticky='news')
    gui.configureWeigths(w.root)
    w.show()


if __name__ == '__main__':
    root = tk.Tk()
    root.withdraw()
    root.title("View files")
    l = TextFileViewer(root, fileList=sys.argv[1:])
    l.pack(side=tk.TOP, fill=tk.BOTH)
    gui.centerWindows(root)
    root.deiconify()
    root.mainloop()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.tests as pwtests
import pyworkflowtests.objects as objectsMod
from pyworkflowtests import Domain


class TestDomain(pwtests.BaseTest):

    def test_objects(self):
        """ Test that all objects are properly discovered. """
        objects = Domain.getObjects()
        for k in dir(objectsMod):
            v = getattr(objectsMod, k)
            if isinstance(v, objectsMod.MockObject):
                self.assertEqual(objects[k], v)

    def test_viewers(self):
        pass

    def test_wizards(self):
        pass



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


"""
This module contains some sqlite basic tools to handle Databases.
"""

import logging
logger = logging.getLogger(__name__)
from sqlite3 import dbapi2 as sqlite
from sqlite3 import OperationalError as OperationalError
from pyworkflow.utils import STATUS, getExtraLogInfo, Config


class SqliteDb:
    """Class to handle a Sqlite database.
    It will create connection, execute queries and commands.
    """
    OPEN_CONNECTIONS = {}  # Store all connections made

    def __init__(self):
        self._reuseConnections = False

    def _createConnection(self, dbName, timeout):
        """Establish db connection"""
        self._dbName = dbName
        if self._reuseConnections and dbName in self.OPEN_CONNECTIONS:
            self.connection = self.OPEN_CONNECTIONS[dbName]
        else:
            # self.closeConnection(dbName)  # Close the connect if exists for this db
            self.connection = sqlite.Connection(dbName, timeout, check_same_thread=False)
            self.connection.row_factory = sqlite.Row
            self.OPEN_CONNECTIONS[dbName] = self.connection
            logger.debug("Connection open for %s" % dbName, extra=getExtraLogInfo(
                "CONNECTIONS",
                STATUS.START,
                dbfilename=dbName))

        self.cursor = self.connection.cursor()
        # Define some shortcuts functions
        if Config.debugSQLOn():
            self.executeCommand = self._debugExecute
        else:
            self.executeCommand = self.cursor.execute
        self.commit = self.connection.commit
        
    @classmethod
    def closeConnection(cls, dbName):
        if dbName in cls.OPEN_CONNECTIONS:
            connection = cls.OPEN_CONNECTIONS[dbName]
            del cls.OPEN_CONNECTIONS[dbName]
            connection.close()
            logger.debug("Connection closed for %s" % dbName,
                         extra=getExtraLogInfo('CONNECTIONS', STATUS.STOP, dbfilename=dbName))

    def getDbName(self):
        return self._dbName
    
    def close(self):
        self.connection.close()
        logger.debug("Connection closed for %s" % self._dbName,
                     extra=getExtraLogInfo(
                                        "CONNECTIONS",
                                        STATUS.STOP,
                                        dbfilename=self._dbName))
        if self._dbName in self.OPEN_CONNECTIONS:
            del self.OPEN_CONNECTIONS[self._dbName]
        
    def _debugExecute(self, *args):
        try:
            logger.debug("COMMAND: %s; %s" %(args[0] , self._dbName),
                extra=getExtraLogInfo("QUERY", STATUS.EVENT, dbfilename=self._dbName)
            )
            logger.debug("ARGUMENTS: " + str(args[1:]))
            return self.cursor.execute(*args)
        except Exception as ex:
            print(">>>> FAILED cursor.execute on db: '%s'" % self._dbName)
            raise ex

    def _iterResults(self):
        row = self.cursor.fetchone()
        while row is not None:
            yield row
            row = self.cursor.fetchone()
        
    def _results(self, iterate=False):
        """ Return the results to which cursor, point to. 
        If iterates=True, iterate yielding each result independently"""
        if not iterate:
            return self.cursor.fetchall()
        else:
            return self._iterResults()
        
    def getTables(self, tablePattern=None):
        """ Return the table names existing in the Database.
        If  tablePattern is not None, only tables matching 
        the pattern will be returned.
        """
        self.executeCommand("SELECT name FROM sqlite_master "
                            "WHERE type='table' "
                            "AND name NOT LIKE 'sqlite_%';")
        return [str(row['name']) for row in self._iterResults()]
    
    def hasTable(self, tableName):
        return tableName in self.getTables()
    
    def getTableColumns(self, tableName):
        self.executeCommand('PRAGMA table_info(%s)' % tableName)
        return self.cursor.fetchall()
    
    def getVersion(self):
        """ Return the database 'version' that is used.
        Internally it make use of the SQLite PRAGMA database.user_version;
        """
        self.executeCommand('PRAGMA user_version')
        return self.cursor.fetchone()[0]
    
    def setVersion(self, version):
        self.executeCommand('PRAGMA user_version=%d' % version)
        self.commit()




    # -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Pablo Conesa [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This modules hosts most of the accessory code that is used in view protocols"""
import json
import os
from configparser import ConfigParser

from pyworkflow import Config
import  pyworkflow.gui as pwgui
import pyworkflow.object as pwobj
import pyworkflow.utils as pwutils
from pyworkflow.gui.project.utils import isAFinalProtocol
from pyworkflow.project import MenuConfig
from pyworkflow.utils import Message, Icon
from pyworkflow.viewer import DESKTOP_TKINTER



class RunIOTreeProvider(pwgui.tree.TreeProvider):
    """Create the tree elements from a Protocol Run input/output children"""

    def __init__(self, parent, protocol, mapper):
        # TreeProvider.__init__(self)
        self.parent = parent
        self.protocol = protocol
        self.mapper = mapper

    @staticmethod
    def getColumns():
        return [('Attribute', 200), ('Info', 100)]

    def getObjects(self):
        objs = []
        if self.protocol is not None:
            # Store a dict with input parents (input, PointerList)
            self.inputParentDict = pwobj.OrderedDict()
            inputs = []
            inputObj = pwobj.String(Message.LABEL_INPUT)
            inputObj._icon = Icon.ACTION_IN
            self.inputParentDict['_input'] = inputObj
            inputParents = [inputObj]

            for key, attr in self.protocol.iterInputAttributes():
                attr._parentKey = key
                # Repeated keys means there are inside a pointerList
                # since the same key is yielded for all items inside
                # so update the parent dict with a new object
                if key in self.inputParentDict:
                    if self.inputParentDict[key] == inputObj:
                        parentObj = pwobj.String(key)
                        parentObj._icon = Icon.ACTION_IN
                        parentObj._parentKey = '_input'
                        inputParents.append(parentObj)
                        self.inputParentDict[key] = parentObj
                else:
                    self.inputParentDict[key] = inputObj
                inputs.append(attr)

            outputs = [attr for _, attr in
                       self.protocol.iterOutputAttributes()]
            self.outputStr = pwobj.String(Message.LABEL_OUTPUT)
            objs = inputParents + inputs + [self.outputStr] + outputs
        return objs

    def _visualizeObject(self, ViewerClass, obj):
        viewer = ViewerClass(project=self.protocol.getProject(),
                             protocol=self.protocol,
                             parent=self.parent.windows)
        viewer.visualize(obj, windows=self.parent.windows)

    def _editObject(self, obj):
        """Open the Edit GUI Form given an instance"""
        pwgui.dialog.EditObjectDialog(self.parent, Message.TITLE_EDIT_OBJECT,
                                      obj, self.mapper)

    def _deleteObject(self, obj):
        """ Remove unnecessary output, specially for Coordinates. """
        prot = self.protocol
        try:
            objLabel = self.getObjectLabel(obj, prot)
            if self.parent.windows.askYesNo("Delete object",
                                            "Are you sure to delete *%s* object?"
                                            % objLabel):
                prot.getProject().deleteProtocolOutput(prot, obj)
                self.parent._fillSummary()
                self.parent.windows.showInfo("Object *%s* successfully deleted."
                                             % objLabel)
        except Exception as ex:
            self.parent.windows.showError(str(ex))

    @staticmethod
    def getObjectPreview(obj):
        desc = "<name>: " + obj.getName()
        return None, desc

    def getObjectActions(self, obj):
        if isinstance(obj, pwobj.Pointer):
            obj = obj.get()
            isPointer = True
        else:
            isPointer = False
        actions = []

        viewers = Config.getDomain().findViewers(obj.getClassName(), DESKTOP_TKINTER)

        def viewerCallback(viewer):
            return lambda: self._visualizeObject(viewer, obj)

        for v in viewers:
            actions.append(('Open with %s' % v.__name__,
                            viewerCallback(v),
                            Icon.ACTION_VISUALIZE))
        # EDIT
        actions.append((Message.LABEL_EDIT,
                        lambda: self._editObject(obj),
                        Icon.ACTION_EDIT))
        # DELETE
        # Special case to allow delete outputCoordinates
        # since we can end up with several outputs and
        # we may want to clean up
        if self.protocol.allowsDelete(obj) and not isPointer:
            actions.append((Message.LABEL_DELETE_ACTION,
                            lambda: self._deleteObject(obj),
                            Icon.ACTION_DELETE))
        return actions

    @staticmethod
    def getObjectLabel(obj, parent):
        """ We will try to show in the list the string representation
        that is more readable for the user to pick the desired object.
        """
        label = 'None'
        if obj:
            label = obj.getObjLabel()
            if not len(label.strip()):
                parentLabel = parent.getObjLabel() if parent else 'None'
                label = "%s -> %s" % (parentLabel, obj.getLastName())
        return label

    def getObjectInfo(self, obj):

        def stringToInfo():
            """ String objects converted to info dictionary for the tree"""

            value = obj.get()
            infoStr = {'key': value, 'text': value, 'values': '', 'open': True}
            if hasattr(obj, '_parentKey'):
                infoStr['parent'] = self.inputParentDict[obj._parentKey]
            return infoStr

        def labelToValue(label, key, name):
            """ To tolerate str(labelObj) in case xmippLib is missing, but
            still being able to open a project."""
            try:
                value = str(label)
            except Exception as e:
                print("Can not convert object %s - %s to string." % (key, name))
                value = str(e)

            return value

        def pointerToInfo():
            """ Converts a Pointer into an info dictionary for the tree"""

            namePtr = obj.getLastName()
            # Remove ugly item notations inside lists
            namePtr = namePtr.replace('__item__000', '')
            # Consider Pointer as inputs
            imagePtr = getattr(obj, '_icon', '')
            parentPtr = self.inputParentDict[obj._parentKey]

            suffix = ''
            if obj.hasExtended():
                # getExtended method remove old attributes conventions.
                extendedValue = obj.getExtended()
                if obj.hasExtended():
                    suffix = '[%s]' % extendedValue
                # else:
                #     suffix = '[Item %s]' % extendedValue

                # Tolerate loading projects:
                # When having only the project sqlite..an obj.get() will
                # the load of the set...and if it is missing this whole
                # "thread" fails.
                try:
                    labelObjPtr = obj.get()
                    if labelObjPtr is None:
                        labelObjPtr = obj.getObjValue()
                        suffix = ''

                except Exception:
                    return {'parent': parentPtr, 'image': imagePtr, 'text': namePtr,
                            'values': ("Couldn't read object attributes.",)}
            else:
                labelObjPtr = obj.get()

            objKeyPtr = obj._parentKey + str(labelObjPtr.getObjId())
            labelPtr = self.getObjectLabel(labelObjPtr,
                                        self.mapper.getParent(labelObjPtr))
            namePtr += '   (from %s %s)' % (labelPtr, suffix)
            valuePtr = labelToValue(labelObjPtr, objKeyPtr, namePtr)
            infoPtr = {'key': objKeyPtr, 'parent': parentPtr, 'image': imagePtr,
                    'text': namePtr, 'values': (valuePtr,)}

            return infoPtr


        if obj is None or not obj.hasValue():
            return None

        if isinstance(obj, pwobj.String):
            info = stringToInfo()
        else:
            # All attributes are considered output, unless they are pointers
            image = Icon.ACTION_OUT
            parent = self.outputStr

            if isinstance(obj, pwobj.Pointer):
                info = pointerToInfo()
            else:
                name = self.getObjectLabel(obj, self.protocol)
                objKey = str(obj.getObjId())
                labelObj = obj
                value = labelToValue(labelObj, objKey, name)
                info = {'key': objKey, 'parent': parent, 'image': image,
                        'text': name, 'values': (value,)}
        return info

class ProtocolTreeConfig:
    """ Handler class that groups functions and constants
    related to the protocols tree configuration.
    """
    ALL_PROTOCOLS = "All"
    TAG_PROTOCOL_DISABLED = 'protocol-disabled'
    TAG_PROTOCOL = 'protocol'
    TAG_SECTION = 'section'
    TAG_PROTOCOL_GROUP = 'protocol_group'
    TAG_PROTOCOL_BETA = 'protocol_beta'
    TAG_PROTOCOL_NEW = 'protocol_new'
    PLUGIN_CONFIG_PROTOCOLS = 'protocols.conf'

    @classmethod
    def getProtocolTag(cls, isInstalled, isBeta=False, isNew=False):
        """ Return the proper tag depending if the protocol is installed or not.
        """
        if isInstalled:
            if isBeta:
                return cls.TAG_PROTOCOL_BETA
            elif isNew:
                return cls.TAG_PROTOCOL_NEW
            return cls.TAG_PROTOCOL
        else:
            return cls.TAG_PROTOCOL_DISABLED

    @classmethod
    def __addToTree(cls, menu, item, checkFunction=None):
        """ Helper function to recursively add items to a menu.
        Add item (a dictionary that can contain more dictionaries) to menu
        If check function is added will use it to check if the value must be added.
        """
        children = item.pop('children', [])

        if checkFunction is not None:
            add = checkFunction(item)
            if not add:
                return
        subMenu = menu.addSubMenu(**item)  # we expect item={'text': ...}
        for child in children:
            cls.__addToTree(subMenu, child, checkFunction)  # add recursively to sub-menu

        return subMenu

    @classmethod
    def __inSubMenu(cls, child, subMenu):
        """
        Return True if child belongs to subMenu
        """
        for ch in subMenu:
            if cls.__isProtocol(child):
                if ch.value is not None and ch.value == child['value']:
                    return ch
            elif ch.text == child['text']:
                return ch
        return None

    @classmethod
    def _orderSubMenu(cls, session):
        """
        Sort all children of a given section:
        The protocols first, then the sections (the 'more' section at the end)
        """

        def sortWhenLastIsAProtocol():
            """ Sorts children when the last is a protocol"""
            for i in range(lastChildPos - 1, -1, -1):
                if childs[i].tag == cls.TAG_PROTOCOL:
                    break
                else:
                    tmp = childs[i + 1]
                    childs[i + 1] = childs[i]
                    childs[i] = tmp

        def sortWhenLastIsNotAProtocol():
            """ Sorts children when the last is NOT a protocol"""
            for i in range(lastChildPos - 1, -1, -1):
                if childs[i].tag == cls.TAG_PROTOCOL:
                    break
                elif 'more' in str(childs[i].text).lower():
                    tmp = childs[i + 1]
                    childs[i + 1] = childs[i]
                    childs[i] = tmp

        lengthSession = len(session.childs)
        if lengthSession > 1:
            childs = session.childs
            lastChildPos = lengthSession - 1
            if childs[lastChildPos].tag == cls.TAG_PROTOCOL:
                sortWhenLastIsAProtocol()
            else:
                sortWhenLastIsNotAProtocol()

    @classmethod
    def __findTreeLocation(cls, subMenu, children, parent):
        """
        Locate the protocol position in the given view
        """
        for child in children:
            sm = cls.__inSubMenu(child, subMenu)
            if sm is None:
                cls.__addToTree(parent, child, cls.__checkItem)
                cls._orderSubMenu(parent)
            elif child['tag'] == cls.TAG_PROTOCOL_GROUP or child['tag'] == cls.TAG_SECTION:
                cls.__findTreeLocation(sm.childs, child['children'], sm)
    @classmethod
    def __isProtocol(cls, dict):
        """ True inf the item has a key named tag with protocol as value"""
        return dict["tag"] == cls.TAG_PROTOCOL

    @classmethod
    def __isProtocolNode(cls, node):
        """ True if tag attribute is protocol"""
        return node.tag == cls.TAG_PROTOCOL


    @classmethod
    def __checkItem(cls, item):
        """ Function to check if the protocol has to be added or not.
        Params:
            item: {"tag": "protocol", "value": "ProtImportMovies",
                   "text": "import movies"}
        """
        if not cls.__isProtocol(item):
            return True

        # It is a protocol as this point, get the class name and
        # check if it is disabled
        protClassName = item["value"]
        protClass = Config.getDomain().getProtocols().get(protClassName)
        icon = 'python_file.gif'
        if protClass is not None:
            if protClass.isBeta():
                icon = "beta.gif"
            elif protClass.isNew():
                icon = "new.gif"
        item['icon'] = icon
        return False if protClass is None else not protClass.isDisabled()

    @classmethod
    def __addAllProtocols(cls, domain, protocols):
        # Add all protocols
        allProts = domain.getProtocols()

        # Sort the dictionary
        allProtsSorted = pwobj.OrderedDict(sorted(allProts.items(),
                                            key=lambda e: e[1].getClassLabel()))
        allProtMenu = ProtocolConfig(cls.ALL_PROTOCOLS)
        packages = {}

        # Group protocols by package name
        for k, v in allProtsSorted.items():
            if isAFinalProtocol(v, k):
                packageName = v.getClassPackageName()
                # Get the package submenu
                packageMenu = packages.get(packageName)

                # If no package menu available
                if packageMenu is None:
                    # Add it to the menu ...
                    packageLine = {"tag": "package", "value": packageName,
                                   "text": packageName}
                    packageMenu = cls.__addToTree(allProtMenu, packageLine)

                    # Store it in the dict
                    packages[packageName] = packageMenu

                # Add the protocol
                tag = cls.getProtocolTag(v.isInstalled(), v.isBeta(), v.isNew())

                protLine = {"tag": tag, "value": k,
                            "text": v.getClassLabel(prependPackageName=False)}

                cls.__addToTree(packageMenu, protLine)

        protocols[cls.ALL_PROTOCOLS] = allProtMenu

    @classmethod
    def __addProtocolsFromConf(cls, protocols, protocolsConfPath):
        """
        Load the protocols in the tree from a given protocols.conf file,
        either the global one in Scipion or defined in a plugin.
        """

        def addProtocols():
            """ Adds protocols defined in the "PROTOCOLS" section of the config file. """
            for menuName in cp.options('PROTOCOLS'):
                if menuName not in protocols:  # The view has not been inserted
                    menu = ProtocolConfig(menuName)
                    children = json.loads(cp.get('PROTOCOLS', menuName))
                    for child in children:
                        cls.__addToTree(menu, child, cls.__checkItem)
                    protocols[menuName] = menu
                else:  # The view has been inserted
                    menu = protocols.get(menuName)
                    children = json.loads(cp.get('PROTOCOLS',
                                                 menuName))
                    cls.__findTreeLocation(menu.childs, children, menu)

        # Populate the protocols menu from the plugin config file.
        if os.path.exists(protocolsConfPath):
            cp = ConfigParser()
            cp.optionxform = str  # keep case
            cp.read(protocolsConfPath)
            #  Ensure that the protocols section exists
            if cp.has_section('PROTOCOLS'):
                addProtocols()

    @classmethod
    def load(cls, domain, protocolsConf):
        """ Read the protocol configuration from a .conf file similar to the
        one in scipion/config/protocols.conf,
        which is the default one when no file is passed.
        """
        protocols = dict()
        # Read the protocols.conf from Scipion (base) and create an initial
        # tree view
        cls.__addProtocolsFromConf(protocols, protocolsConf)

        # Read the protocols.conf of any installed plugin
        pluginDict = domain.getPlugins()
        pluginList = cls.__orderByPriority(pluginDict.keys(),
                                           priorityPluginList=Config.getPriorityPackageList())
        for pluginName in pluginList:
            try:

                # if the plugin has a path
                if pwutils.isModuleLoaded(pluginName) and pwutils.isModuleAFolder(pluginName):
                    # Locate the plugin protocols.conf file
                    protocolsConfPath = os.path.join(
                        pluginDict[pluginName].__path__[0],
                        cls.PLUGIN_CONFIG_PROTOCOLS)
                    cls.__addProtocolsFromConf(protocols, protocolsConfPath)

            except Exception as e:
                print('Failed to read settings. The reported error was:\n  %s\n'
                      'To solve it, fix %s and run again.' % (
                          e, os.path.abspath(protocolsConfPath)))

        # Clean empty sections
        cls._hideEmptySections(protocols)

        # Add all protocols to All view
        cls.__addAllProtocols(domain, protocols)

        return protocols

    @classmethod
    def _hideEmptySections(cls, protocols):
        """ Cleans all empty sections in the tree"""

        for protConf in protocols.values():
            cls._setVisibility(protConf)

    @classmethod
    def _setVisibility(cls, node):
        """ Sets the visibility of a node based on the presence of a leaf hanging form it"""
        if cls.__isProtocolNode(node):
            # Default visibility value is true. No need to set it again
            return True

        anyLeaf = False

        for child in node.childs:
            # NOTE: since python short circuits this, _setVisibility must be called always. So not swap!!
            anyLeaf = cls._setVisibility(child) or anyLeaf

        node.visible = anyLeaf

        return anyLeaf
    @classmethod
    def __orderByPriority(cls, pluginList, priorityPluginList):
        if priorityPluginList:
            sortedPluginList = priorityPluginList + [pluginName for pluginName in pluginList
                                                     if pluginName not in priorityPluginList]
            return sortedPluginList
        else:
            return pluginList


class ProtocolConfig(MenuConfig):
    """Store protocols configuration """

    def __init__(self, text=None, value=None, **args):
        MenuConfig.__init__(self, text, value, **args)
        if 'openItem' not in args:
            self.openItem = self.tag != 'protocol_base'

    def addSubMenu(self, text, value=None, shortCut=None, **args):
        if 'icon' not in args:
            tag = args.get('tag', None)
            if tag == 'protocol_base':
                args['icon'] = 'class_obj.gif'

        args['shortCut'] = shortCut
        return MenuConfig.addSubMenu(self, text, value, **args)

    def __str__(self):
        return self.text



import time
instances = {}
elapsedTime = {}


def getInstancesCount():
    return instances


def instanceCreated(cls):
    instances[cls] = instances.get(cls, 0) + 1


def instanceDestroyed(cls):
    instances[cls] = instances.get(cls, 0) - 1


def printInstances():
    for k, v in instances.items():
        print("%s: %d", k, v)


def waitForDebugger(seconds=20):
    print("Waiting for debugger %d seconds." % seconds)
    from pyworkflow.utils import printTraceBack
    printTraceBack()

    while seconds > 0:
        time.sleep(1)
        # Set seconds to 0:
        # Execute this in the debugger:  seconds = 0
        print(str(seconds) + " seconds left.")
        seconds -= 1


# From: https://www.huyng.com/posts/python-performance-analysis


class Timer(object):
    indentation = 0
    blackList = []
    whiteList = []
    minms = 1

    def __init__(self, name=None, verbose=True):
        self.verbose = verbose
        self.name = name
        Timer.indentation += 1

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.secs = self.end - self.start
        self.msecs = self.secs * 1000  # millisecs
        Timer.indentation -= 1

        toPrint = (len(Timer.whiteList) == 0)

        if any(self.name.startswith(s) for s in Timer.blackList):
            toPrint = False

        if any(self.name.startswith(s) for s in Timer.whiteList):
            toPrint = True

        if self.verbose and toPrint and self.msecs > Timer.minms:
            print('{0}{1:10.0f}\tms\t{2}'.format('\t' * Timer.indentation, self.msecs, self.name))

# To monitor memory leaks:
# 1.- Check you have installed pympler: scipion run pip install pympler
# 2.- Uncomment lines below
# 3.- Add a with statement like:
#     with monitor.MemoryMonitor():
#         <code to analyze>
# from pympler.tracker import SummaryTracker
#
#
# class MemoryMonitor(object):
#
#     def __init__(self):
#         self.tracker = None
#
#     def __enter__(self):
#         self.tracker = SummaryTracker()
#         self.tracker.print_diff()
#         return self
#
#     def __exit__(self, *args):
#         self.tracker.print_diff()
